import ctypes
import sys
import os




class JuliaModuleInfo(ctypes.Structure):
    _fields_ = [("setup", ctypes.c_voidp),
                ("execute", ctypes.c_voidp),
                ("cleanup", ctypes.c_voidp),
                ("name", ctypes.c_char_p),
                ]

JuliaModuleInfoPtr = ctypes.POINTER(JuliaModuleInfo)

class JuliaModule(object):
    lib = None
    def __init__(self, filepath):
        dirname, filename = os.path.split(filepath)
        module_name = filename[:-3]
        self.setup_library()
        print("Loading module")
        self.info = self.lib.load_module(dirname.encode('ascii'), module_name.encode('ascii'))
        if not self.info:
            raise ValueError("Error loading Julia module at {}".format(filepath))

    @classmethod
    def setup_library(cls):
        if cls.lib is not None:
            return

        libdir = os.path.split(__file__)[0]
        libname = os.path.join(libdir, "libcosmosis_julia.so")

        if not os.path.exists(libname):
            raise RuntimeError("You must compile cosmosis with Julia support by running make in {}".format(libdir))
        cls.lib = ctypes.CDLL(libname, mode=ctypes.RTLD_GLOBAL)

        cls.lib.load_module.restype = JuliaModuleInfoPtr
        cls.lib.load_module.argtypes = [ctypes.c_char_p, ctypes.c_char_p]

        cls.lib.run_setup.argtypes = [JuliaModuleInfoPtr, ctypes.c_voidp]
        cls.lib.run_setup.restype = ctypes.c_voidp

        cls.lib.run_execute.argtypes = [JuliaModuleInfoPtr, ctypes.c_voidp, ctypes.c_voidp]
        cls.lib.run_execute.restype = ctypes.c_int

        cls.lib.run_cleanup.argtypes = [JuliaModuleInfoPtr, ctypes.c_voidp]
        cls.lib.run_cleanup.restype = ctypes.c_int


    def setup(self, options_ptr):
        config = self.lib.run_setup(self.info, options_ptr)
        if not config:
            raise ValueError("Error running Julia setup function - backtrace above")
        return config

    def execute(self, block_ptr, config_ptr):
        return self.lib.run_execute(self.info, block_ptr, config_ptr)

    def cleanup(self, config_ptr):
        return self.lib.run_cleanup(self.info, config_ptr)




from cosmosis.runtime import register_new_parameter

def setup(options):
    # example of how to register a new parameter in the pipeline
    # from where.
    register_new_parameter(
        options, # the options datablock we were just passed
        "new_parameters", # section
        "p3",  # name
        -1.0,  # min value
        0.0,   # start value
        1.0,   # max value
        "normal",  # prior name, optional, defaults to uniform or delta if min==max.  see prior.py
        [0.1, 0.2] # prior parameters, optional, in this case [mu, sigma], see prior.py
    )
    register_new_parameter(
        options, # the options datablock we were just passed
        "new_parameters", # section
        "delta",  # name
        1.0,  # min value
        1.0,   # start value
        1.0,   # max value
    )
    return {}

def execute(block, config):
    p1 = block['parameters', 'p1']
    p2 = block['parameters', 'p2']
    p3 = block['new_parameters', 'p3']
    delta = block['new_parameters', 'delta']
    assert delta == 1.0
    return 0



#coding: utf-8


u"""Definition of `everythingisnan`, :class:`ParseExtraParameters`, :func:`mkdir`, :class:`Timer`, :func:`symmetrized_matrix` and :func:`symmetric_positive_definite_inverse`."""


import numpy as np
import argparse
import os
import errno
from timeit import default_timer
import sys
from contextlib import contextmanager
import tempfile
import subprocess
from functools import wraps
import pathlib
try:
    import dulwich
except:
    dulwich = None


class EverythingIsNan:

    u"""An object which, when iterated over or indexed directly, always returns NumPyʼs `np.nan` as a value."""
    
    def __getitem__(self, param):
        u"""Just return NaN."""
        return np.nan


everythingIsNan = EverythingIsNan()



def datablock_to_astropy(block):
    from .datablock import names
    cosmo = names.cosmological_parameters
    from astropy.cosmology import FlatLambdaCDM, LambdaCDM, FlatwCDM, wCDM, w0waCDM, Flatw0waCDM

    # Required parameters that must be in the data block
    # for this to work
    omega_lambda = block[cosmo, 'omega_lambda']
    omega_m = block[cosmo, 'omega_m']
    omega_b = block[cosmo, 'omega_b']
    H0 = block[cosmo, 'h0'] * 100
    m_nu = block[cosmo, 'mnu']

    # Optional parameters with simple default values
    omega_k = block.get_double(cosmo, 'omega_k', 0.0)
    w0 = block.get_double(cosmo, 'w', -1.0)
    wa = block.get_double(cosmo, 'wa', 0.0)

    kw = {
        "H0": H0,
        "Ob0": omega_b,
        "Om0": omega_m,
        "m_nu": m_nu,
    }

    if omega_k == 0:
        if wa == 0:
            if w0 == -1.0:
                model = FlatLambdaCDM
            else:
                model = FlatwCDM
                kw["w0"] = w0
        else:
            model = Flatw0waCDM
            kw["w0"] = w0
            kw["wa"] = wa
    else:
        if wa == 0:
            if w0 == -1.0:
                model = LambdaCDM
                kw["Ode0"] = omega_lambda
            else:
                model = wCDM
                kw["Ode0"] = omega_lambda
                kw["w0"] = w0
        else:
            model = w0waCDM
            kw["Ode0"] = omega_lambda
            kw["w0"] = w0
            kw["wa"] = wa

    return model(**kw)



class ParseExtraParameters(argparse.Action):

    u"""Extended command-line argument parser :class:`Action` which knows how to read arguments of the form ‘section.name=value’."""

    def __call__(self, parser, args, values, option_string=None):
        if getattr(args, self.dest, self.default) is not None:
            parser.error(option_string + " appears several times")
        result = {}
        for arg in values:
            section, param_value = arg.split('.',1)
            param,value = param_value.split('=',1)
            result[(section,param)] = value
        setattr(args, self.dest, result)



def mkdir(path):
    u"""Ensure that all the components in the `path` exist in the file system.

    If there is a file in the file system blocking the creation of a
    directory there, a :class:`ValueError` will be raised.  Any other
    problem will raise an underlying `os` exception.

    """
    #This is much nicer in python 3.
    # Avoid trying to make an empty path
    if not path.strip():
        return
    try:
        os.makedirs(path)
    except OSError as error:
        if error.errno == errno.EEXIST:
            if os.path.isdir(path):
                #error is that dir already exists; fine - no error
                pass
            else:
                #error is that file with name of dir exists already
                raise ValueError("Tried to create dir %s but file with name exists already"%path)
        elif error.errno == errno.ENOTDIR:
            #some part of the path (not the end) already exists as a file 
            raise ValueError("Tried to create dir %s but some part of the path already exists as a file"%path)
        else:
            #Some other kind of error making directory
            raise


class Timer(object):
    u"""Object to be use with `with` instruction, so that when enclosed code completes a message will appear with the elapsed wall-clock time."""
    def __init__(self, msg):
        u"""Set `msg` as a label which identifies the code block being timed."""
        self.msg = msg

    def __enter__(self):
        u"""Start the timer running."""
        self.start = default_timer()
        return self

    def __exit__(self, *args):
        u"""Print out the elapsed time."""
        interval = default_timer() - self.start
        print("Time taken by step '{}': {}".format(self.msg, interval))



def symmetrized_matrix(U):
    u"""Return a new matrix like `U`, but with upper-triangle elements copied to lower-triangle ones."""
    M = U.copy()
    inds = np.triu_indices_from(M,k=1)
    M[(inds[1], inds[0])] = M[inds]
    return M



def symmetric_positive_definite_inverse(M):
    u"""Compute the inverse of a symmetric positive definite matrix `M`.

    A :class:`ValueError` will be thrown if the computation cannot be
    completed.

    """
    import scipy.linalg
    U,status = scipy.linalg.lapack.dpotrf(M)
    if status != 0:
        raise ValueError("Non-symmetric positive definite matrix")
    M,status = scipy.linalg.lapack.dpotri(U)
    if status != 0:
        raise ValueError("Error in Cholesky factorization")
    M = symmetrized_matrix(M)
    return M

# These parts from:
# https://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python

def fileno(file_or_fd):
    fd = getattr(file_or_fd, 'fileno', lambda: file_or_fd)()
    if not isinstance(fd, int):
        raise ValueError("Expected a file (`.fileno()`) or a file descriptor")
    return fd

@contextmanager
def stdout_redirected(to=os.devnull, stdout=None):
    from .runtime import logs
    if stdout is None:
       stdout = sys.stdout

    stdout_fd = fileno(stdout)
    # copy stdout_fd before it is overwritten
    #NOTE: `copied` is inheritable on Windows when duplicating a standard stream
    with os.fdopen(os.dup(stdout_fd), 'wb') as copied: 
        stdout.flush()  # flush library buffers that dup2 knows nothing about
        log_level = logs.get_level()
        logs.set_level(51)
        try:
            os.dup2(fileno(to), stdout_fd)  # $ exec >&to
        except ValueError:  # filename
            with open(to, 'wb') as to_file:
                os.dup2(to_file.fileno(), stdout_fd)  # $ exec > to
        try:
            yield stdout # allow code to be run with the redirected stdout
        finally:
            # restore stdout to its previous value
            #NOTE: dup2 makes stdout_fd inheritable unconditionally
            stdout.flush()
            os.dup2(copied.fileno(), stdout_fd)  # $ exec >&copied
            logs.set_level(log_level)








def read_comment_section(filename):
    """Pull out the comments section from the top of a chain file"""
    lines = []
    for line in open(filename):
        if not line.startswith('#'):
            break
        lines.append(line)
    return lines


def extract_section(lines, section):
    """Extract the PARAMS, VALUES, or PRIORS section from a group
    of comment lines from a chain file"""
    start = "## START_OF_{}_INI".format(section).upper()
    end = "## END_OF_{}_INI".format(section).upper()
    in_section = False
    output_lines = []
    for line in lines:
        if line.startswith(start):
            in_section = True
            continue
        elif line.startswith(end):
            break
        elif in_section:
            output_lines.append(line[3:])
    return output_lines

def save_section(lines, section, prefix):
    """Save a group of lines to a file"""
    filename = "{}_{}.ini".format(prefix, section)
    open(filename,'w').writelines(lines)

def extract_params(chain, prefix):
    """ Extract the parameters, values, and priors files from file "chain"
    and save them to prefix_params.ini etc. """
    lines = read_comment_section(chain)

    for section in ['params', 'values', 'priors']:
        section_lines = extract_section(lines, section)
        save_section(section_lines, section, prefix)


def tempdir_safe_docker():
    # Under python3 TemporaryDirectory is in the tempfile
    # standard library package.
    # Under python2 the backports.tempfile package is required.
    # It can be pip installed.
    try:
        TemporaryDirectory = tempfile.TemporaryDirectory
    except AttributeError:
        try:
            import backports.tempfile
            TemporaryDirectory = backports.tempfile.TemporaryDirectory
        except ImportError:
            raise ImportError("In python 2 the PriorFunction code "+
                "requires you to install backports.tempfile.  It can be done with pip")

    return TemporaryDirectory()


class PriorFunction(object):
    def __init__(self, chain_filename):
        "Build a prior function from an input chain file"
        import cosmosis.runtime.parameter


        # Read the parameter files to temporary files
        # they will be deleted afterwards
        with tempdir_safe_docker() as tmpdir:
            dirname = tmpdir + os.path.sep
            extract_params(chain_filename, dirname+"tmp")

            # Load in the extracter parameters, including their priors
            self.all_params = cosmosis.runtime.parameter.Parameter.load_parameters(
                dirname+'tmp_values.ini',
                priors_files=[dirname+'tmp_priors.ini'])

        # Pull out the priors from the parameters
        self.all_priors = [p.prior for p in self.all_params]
        self.varied_priors = [p.prior for p in self.all_params if p.is_varied()]

    def _evaluate(self, p_in, priors):
        # Internal method
        # Convert to 2D array
        p = np.atleast_2d(p_in)

        #Check shape of array
        if not p.shape[1]==len(priors):
            raise ValueError("Wrong dimension in evaluate_all")
        # Number of samples
        n = p.shape[0]
        # Output value
        logp = np.zeros(n)
        # Loop through samples and then through parameters
        # in that sample.
        for i,p_i in enumerate(p):
            logp_i = 0.0
            for x, prior in zip(p_i, priors):
                logp_i += prior(x)  # Calling a prior returns log(P(x))
            logp[i] = logp_i
        return logp

    def evaluate_p(self, p_in):
        """
        Evaluate P(x) of the varied parameters.
        p_in is a 1D array if length nparam
        or a 2D array of shape (nsample, nparam)
        """
        return np.exp(self.evaluate_logp(p_in))

    def evaluate_p_all(self, p_in):
        """
        Evaluate log(P(x)) of all the parameters, including fixed ones.
        p_in is a 1D array if length nparam
        or a 2D array of shape (nsample, nparam)
        """
        return np.exp(self.evaluate_logp_all(p_in))

    def evaluate_logp(self, p_in):
        """
        Evaluate log(P(x)) of the varied parameters.
        p_in is a 1D array if length nparam
        or a 2D array of shape (nsample, nparam)
        """
        return self._evaluate(p_in, self.varied_priors)

    def evaluate_logp_all(self, p_in):
        """
        Evaluate log(P(x)) of all the parameters, including fixed ones.
        p_in is a 1D array if length nparam
        or a 2D array of shape (nsample, nparam)
        """
        return self._evaluate(p_in, self.all_priors)

def get_git_revision_dulwich(directory):
    import dulwich.repo
    path = pathlib.Path(directory)
    while not (path / ".git").exists():
        if str(path) == path.root:
            # not a git repo
            return ""
        path = path.parent.absolute()
    repo = dulwich.repo.Repo(path)
    return repo.head().decode('utf-8')


def get_git_revision(directory):
    try:
        # Turn e.g. $COSMOSIS_SRC_DIR into a proper path
        directory = os.path.expandvars(directory)
        if not os.path.isdir(directory):
            return ""

        if dulwich is not None:
            return get_git_revision_dulwich(directory)

        if os.environ.get("COSMOSIS_NO_SUBPROCESS", "") not in ["", "0"]:
            return ""

        # this git command gives the current commit ID of the
        # directory it is run from
        cmd = "git rev-parse HEAD".split()

        # run, capturing stderr and stdout to read the hash from
        p = subprocess.Popen(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
            cwd=directory)

        # Read stdout.  Discard stderr, which will be None
        rev, _ = p.communicate()

        # If there are any errors then we ignore everything
        if p.returncode:
            return ""
        # There shouldn't be any newlines here, but in case there are in future
        # we replace them with spaces to avoid messing up output file formats
        return rev.decode('utf-8').strip().replace("\n", " ")
    
    # It's usually bad practice to catch all exceptions, but we really don't
    # want to crash if we can't get the metadata.
    except Exception as e:
        print(f"Unable to get metadata on git revision for {directory}: error {e}")
        return ""



def import_by_path(name, path):
    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    import importlib.util
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def underline(s, char='-'):
    """
    Return a string with a line of the given character
    underneath it.

    Parameters
    ----------
    s : str
        The string to underline
    char : str, optional
        The character to use for the line
    """
    return s + "\n" + char * len(s)

def overline(s, char='-'):
    """
    Return a string with a line of the given character
    over it.

    Parameters
    ----------
    s : str
        The string to overline
    char : str, optional
        The character to use for the line
    """
    return char * len(s) + "\n" + s

def under_over_line(s, char='-'):
    """
    Return a string with a line of the given character
    over and under it.

    Parameters
    ----------
    s : str
        The string to overline
    char : str, optional
        The character to use for the line
    """
    return underline(overline(s, char), char)


def read_chain_header(filename):
    lines = []
    for line in open(filename):
        if not line.startswith('#'):
            break
        lines.append(line)
    return lines


def extract_inis_from_chain_header(lines, section):
    start = "## START_OF_{}_INI".format(section).upper()
    end = "## END_OF_{}_INI".format(section).upper()
    in_section = False
    output_lines = []
    for line in lines:
        if line.startswith(start):
            in_section = True
            continue
        elif line.startswith(end):
            break
        elif in_section:
            output_lines.append(line[3:])
    return output_lines


from .. import ParallelSampler
try:
    import abcpmc
    from abcpmc import mpi_util
except ImportError:
    raise ValueError("To use ABC PMC you need to install it with pip install abcpmc")
import numpy as np
from ...runtime import prior

def abc_model(p):
    data = abc_pipeline.run_parameters(p)
    if data is None:
        return None
    try:
        model = data[names.data_vector, 'simulation']
    except BlockError:
        raise ValueError("The module in the AB pipeline should save a model with the same dimensions as the data") 
    return model


class ABCSampler(ParallelSampler):
    parallel_output = False #True for 
    sampler_outputs = [("weight", float),("distance",float)]

    def config(self):
        global abc_pipeline
        abc_pipeline = self.pipeline


        if self.is_master():
            if self.pool is not None:
                raise ValueError("mpi ABC PMC not supported yet... ")
                self.pool.close()
            self.abcpmc = abcpmc
            self.threshold = self.read_ini("threshold",str, 'LinearEps')
            self.epimax = self.read_ini('epimax', float,5.0)
            self.epimin = self.read_ini('epimin',float, 1.0)
            self.part_prop = self.read_ini("particle_prop",str,'weighted_cov')
            self.set_prior = self.read_ini("set_prior",str,'Gaussian')
            self.knn = self.read_ini("num_nn",int, 10)
            self.npart = self.read_ini("npart",int,100)
            self.niter = self.read_ini("niter",int,2)
            self.ngauss = self.read_ini("ngauss",int,4)
            self.run_multigauss = self.read_ini("run_multigauss",bool,False)
            self.ndim = len(self.pipeline.varied_params)

            #options for decreasing threshold
            if self.threshold == 'ConstEps':
                self.eps = self.abcpmc.ConstEps(self.niter, self.epimax)
            elif self.threshold == 'ExpEps':
                self.eps = self.abcpmc.ExponentialEps(self.niter, self.epimax,self.epimin)
            else:
                self.eps = self.abcpmc.LinearEps(self.niter, self.epimax, self.epimin)

            print("\nRunning ABC PMC")
            print("with %d particles, %s prior, %s threshold, %d iterations over (%f,%f), %s kernal \n" % (self.npart,self.set_prior,self.threshold,self.niter,self.epimax,self.epimin,self.part_prop))

            #Data file is read for use in dist() for each step
            #parameter covariance used in the prior   
            if self.run_multigauss:
                self.sigma,self.data = self.generate_data(5000,0.25,1.)
            else:
                self.data = self.load_data()
                self.sigma = self.load_covariance_matrix()
        
            #At the moment the same prior (with variable hyperparameters) is
            # used for all parameters  - would be nice to change this to be more flexible 
            self.pmin = np.zeros(self.ndim)
            self.pmax = np.zeros(self.ndim)
            for i,pi in enumerate(self.pipeline.varied_params):
                self.pmin[i] = pi.limits[0]
                self.pmax[i] = pi.limits[1]
            self.p0 = np.array([param.start for param in self.pipeline.varied_params])
            if self.set_prior == 'uniform':
                self.prior = self.abcpmc.TophatPrior(self.pmin,self.pmax)
            else:
                self.prior = self.abcpmc.GaussianPrior(self.p0, self.sigma*2) 

            #create sampler
            self.sampler = self.abcpmc.Sampler(N=self.npart, Y=self.data, postfn=abc_model, dist=self.dist,pool=self.pool)

            #set particle proposal kernal
            self.abcpmc.Sampler.particle_proposal_kwargs = {}
            if self.part_prop == 'KNN':
                self.abcpmc.Sampler.particle_proposal_kwargs = {'k':self.knn}
                self.sampler.particle_proposal_cls = self.abcpmc.KNNParticleProposal
            elif self.part_prop == 'OLCM':
                self.sampler.particle_proposal_cls = self.abcpmc.OLCMParticleProposal

            self.converged = False

    def output_samples(self,out):
        for o in out:
            self.output.parameters(o)

    def execute(self):
        # called only by master. Other processors will be in self.pool
        # generate outputs somehow by running log_probability_function
        outputs = []
        for output in self.sampler.sample(self.prior, self.eps):
            if self.run_multigauss:
                print("T: {0}, eps: {1:>.4f}, ratio: {2:>.4f}".format(output.t, output.eps, output.ratio))
                for i, (mean, std) in enumerate(zip(np.mean(output.thetas, axis=0), np.std(output.thetas, axis=0))):
                    print(u"    theta[{0}]: {1:>.4f} \u00B1 {2:>.4f}".format(i, mean,std))
            outputs.append(np.c_[output.thetas,output.ws,output.dists])
        for out in outputs:
            self.output_samples(out)

        self.converged = True


    def is_converged(self):
        return  self.converged

     #distance function: sum of abs mean differences
        #Could add other distance functions here -EJ
    def dist(self,x, y):
        return np.sum(np.abs(np.mean(x, axis=0) - np.mean(y, axis=0)))


    def load_covariance_matrix(self):
        covmat_filename = self.read_ini("covmat", str, "").strip()
        if covmat_filename == "":
            covmat = np.array([p.width()/100.0 for p in self.pipeline.varied_params])
        elif not os.path.exists(covmat_filename):
            raise ValueError(
            "Covariance matrix %s not found" % covmat_filename)
        else:
            covmat = np.loadtxt(covmat_filename)

        if covmat.ndim == 0:
            covmat = covmat.reshape((1, 1))
        elif covmat.ndim == 1:
            covmat = np.diag(covmat ** 2)

        nparams = len(self.pipeline.varied_params)
        if covmat.shape != (nparams, nparams):
            raise ValueError("The covariance matrix was shape (%d x %d), "
                    "but there are %d varied parameters." %
                    (covmat.shape[0], covmat.shape[1], nparams))
        return covmat

    def load_data(self):
        data_filename = self.read_ini("datafile", str, "").strip()
        if not os.path.exists(data_filename):
            raise ValueError("Data file %s not found" % data_filename)
        else:
            self.data = np.loadtxt(data_filename)

    def generate_data(self,size,std,Max):
        sigma = np.eye(self.ngauss) * std
        means = Max*np.random.random_sample((self.ngauss,))
        if self.run_multigauss: print("True means: " ,means)
        np.savetxt("abc_multigauss_means.txt",np.vstack(means))
        data = np.random.multivariate_normal(means, sigma, size)
        return sigma,data



import numpy as np


class Proposal(object):
    def __init__(self, cholesky, scaling=2.4, exponential_probability=0.333333):
        self.iteration = 0
        self.ndim = len(cholesky)
        self.cholesky = cholesky 
        rotation = np.identity(self.ndim)
        self.proposal_axes = np.dot(self.cholesky, rotation)
        self.scaling = scaling
        self.exponential_probability = exponential_probability

    def proposal_distance(self, ndim, scaling):
        #from CosmoMC
        if np.random.uniform()<self.exponential_probability:
            r = np.random.exponential()
        else:
            n = min(ndim,2)
            r = (np.random.normal(size=n)**2).mean()**0.5
        return r * scaling

    def randomize_rotation(self):
        #After CosmoMC, we randomly rotate our proposal axes
        #to avoid doubling back on ourselves
        rotation = random_rotation_matrix(self.ndim)
        #All our proposals are done along the axes of our covariance
        #matrix
        self.proposal_axes = np.dot(self.cholesky, rotation)

    def propose(self, p):
        #Once we have cycled through our axes, re-randomize them
        i = self.iteration%self.ndim
        if i==0:
            self.randomize_rotation()
        self.iteration += 1
        #otherwise, propose along our defined axes
        return p + self.proposal_distance(self.ndim,self.scaling) * self.proposal_axes[:,i]



class FastSlowProposal(Proposal):
    def __init__(self, covariance, fast_indices, slow_indices, oversampling, scaling=2.4, exponential_probability=0.3333):
        self.ordering = np.concatenate([slow_indices, fast_indices])
        self.inverse_ordering = invert_ordering(self.ordering)
        self.nslow = len(slow_indices)
        self.oversampling = oversampling
        self.iteration = 0
        self.slow_iteration = 0
        reordered_covariance = covariance[:,self.ordering][self.ordering]
        reordered_cholesky = np.linalg.cholesky(reordered_covariance)

        #For the fast subspace we just use the original vanilla proposal.
        #The slow proposal must be a little different  - not a square matrix
        self.fast_proposal = Proposal(reordered_cholesky[self.nslow:, self.nslow:], exponential_probability=exponential_probability)
        self.slow_matrix = reordered_cholesky[:,:self.nslow]

        self.slow_rotation = np.identity(self.nslow)
        self.scaling = scaling
        self.exponential_probability = exponential_probability



    def propose(self, p):
        p = p[self.ordering]
        if self.iteration%(self.oversampling+1)==0:
            q = self.propose_slow(p)
        else:
            q = self.propose_fast(p)
        self.iteration += 1
        return q[self.inverse_ordering]

    def propose_fast(self, p):
        q = np.zeros_like(p)
        q[:self.nslow] = p[:self.nslow]
        q[self.nslow:] += self.fast_proposal.propose(p[self.nslow:])
        return q
        
    def propose_slow(self, p):
        i = self.slow_iteration%self.nslow
        if i==0:
            self.randomize_rotation()   

        #Following the notation in Lewis (2013)
        delta_s = self.slow_rotation[i] * self.proposal_distance(self.nslow,self.scaling)
        self.slow_iteration += 1

        q = p + np.dot(self.slow_matrix, delta_s)
        return q
        

    def randomize_rotation(self):
        self.slow_rotation = random_rotation_matrix(self.nslow)



#I've been copying this algorithm out of CosmoMC
#for the last decade.  Every time I need a new one
#I think to myself that this time I'll heed the warning
#in the CosmoMC code that, quote:
#!this is most certainly not the world's most efficient or 
#robust random rotation generator"
#unquote, and that I'll try and dig out a better one.
#And every time I spend about an hour looking, before
#coming back to this and translating it.
def random_rotation_matrix(n):
    R=np.identity(n)
    for j in range(n):
        while True:
            v = np.random.normal(size=n)
            for i in range(j):
                v -= R[i,:] * np.dot(v,R[i,:])
            L = np.dot(v,v)
            if (L>1e-3): break
        R[j,:] = v/L**0.5
    return R



def submatrix(M, x):
    """If x is an array of integer row/col numbers and M a matrix,
    extract the submatrix which is the all x'th rows and cols.
    i.e. A = submatrix(M,x) => A_ij = M_{x_i}{x_j}
    """
    return M[np.ix_(x,x)]


def invert_ordering(ordering):
    n = len(ordering)
    inverse = np.zeros_like(ordering)
    for i,j in enumerate(ordering):
        inverse[j] = i
    return inverse



from cosmosis.output.text_output import TextColumnOutput
from cosmosis.output.cosmomc_output import CosmoMCOutput
import tempfile
import string
import numpy as np
import os
try:
    import astropy.table
except:
    astropy = None

def populate_table(out, nparam, ns):
    out.metadata('NP',nparam)
    out.metadata('NS',ns)
    out.metadata('TIME','1:30pm')

    for i in range(nparam):
        p = string.ascii_uppercase[i]
        out.add_column(p, float, 'The parameter called %s'%p)

    for i in range(ns):
        x = np.arange(nparam, dtype=int)+i
        out.parameters(x)
    out.final("FINISH",True)
    out.close()

def test_text():
    with tempfile.TemporaryDirectory() as dirname:
        filename=os.path.join(dirname, 'cosmosis_temp_output_test.txt')
        ini = {'filename':filename, 'format':'text'}
        out = TextColumnOutput.from_options(ini)
        nparam = 8
        ns = 20
        populate_table(out, nparam, ns)

        #We should be able to load this table with loadtxt
        if astropy:
            t = astropy.table.Table.read(filename, format='ascii.commented_header')
            A = t['A']
            B = t['B']
        else:
            t = np.loadtxt(filename,dtype=int).T
            A = t[0]
            B = t[1]

        assert (A == np.arange(ns, dtype=int)).all()
        assert (B == np.arange(ns, dtype=int)+1).all()


        #or with our own method
        names, data, meta, comments, final = TextColumnOutput.load_from_options({"filename":filename})
        assert names == [string.ascii_uppercase[i] for i in range(nparam)]
        assert len(names)==nparam
        assert len(data[0])==ns
        assert meta[0]['NP']==nparam
        assert final[0]['FINISH'] is True

def test_cosmomc_output():
    with tempfile.TemporaryDirectory() as dirname:
        filename=os.path.join(dirname, 'cosmosis_temp_cosmomc_output_test.txt')

        ini = {'filename':filename, 'format':'cosmomc'}
        out = CosmoMCOutput.from_options(ini)

        nparam = 8
        ns = 20
        for i in range(nparam):
            p = string.ascii_uppercase[i]
            out.add_column(p, float, 'The parameter called %s'%p)
        out.add_column("post", float, 'The parameter called %s'%p)


        for i in range(ns):
            x = np.arange(nparam + 1, dtype=int)+i
            x[-1] = -666.0 + i
            out.parameters(x)

        # do the last one again
        out.parameters(x)


        # Need to call this to print out the last one
        out._close()


        data = np.loadtxt(filename)
        assert data.shape == (ns, nparam + 2)
        assert (data[:-1, 0] == 1).all()
        assert data[-1, 0] == 2



import numpy as np

class Hints(object):
    def __init__(self):
        self._peak = None
        self._cov = None
        self._peak_post = None

    def set_from_sample(self, samples, posts, weights=None, log_weights=None):
        assert len(samples) == len(posts)
        self.set_peak_from_sample(samples, posts)
        if log_weights is not None:
            if weights is not  None:
                raise ValueError("You must provide either weights or log_weights, not both")
            weights = np.exp(log_weights - log_weights.max())
        self.set_cov(np.cov(samples.T, aweights=weights))

    def has_peak(self):
        return self._peak is not None
    def set_peak(self, peak, post):
        if self._peak_post is None or post > self._peak_post:
            self._peak = np.array(peak)
            self._peak_post = post

    def set_peak_from_sample(self, samples, posts):
        assert len(samples) == len(posts)
        idx = np.argmax(posts)
        self.set_peak(samples[idx], posts[idx])
    def get_peak(self):
        return self._peak
    def del_peak(self):
        self._peak = None
        self._peak_post = None
    def has_cov(self):
        return self._cov is not None
    def set_cov(self, cov):
        self._cov = cov
    def get_cov(self):
        return self._cov
    def del_cov(self):
        self._cov = None
    
    def update(self, other):
        if other.has_peak():
            self.set_peak(other.get_peak(), other._peak_post)
        if other.has_cov():
            self.set_cov(other.get_cov())


from .. import ParallelSampler
import numpy as np



neval = 0
nfail = 0
def minus_log_posterior(p_in):
    global neval, nfail
    p_in, index = p_in
    if maxlike_sampler.pool:
        rank = maxlike_sampler.pool.rank
    else:
        rank = 0
    if (not np.all(p_in>=0)) or (not np.all(p_in<=1.0)):
        nfail += 1
        pstr = '   '.join(str(x) for x in p_in)
        return np.inf
    neval +=1 
    p = maxlike_sampler.pipeline.denormalize_vector(p_in)
    post, extra = maxlike_sampler.pipeline.posterior(p)
    pstr = '   '.join(str(x) for x in p)
    return -post



def posterior_and_gradient(p_in):
    pstr = '   '.join(str(x) for x in p_in)
    print("Calculating gradient about (normalized) point {}".format(pstr))
    points = [(p_in,0)]
    n = len(p_in)
    for i in range(n):
        p = p_in.copy()
        p[i] += maxlike_sampler.epsilon
        points.append((p, i+1))

    if maxlike_sampler.pool:
        results = maxlike_sampler.pool.map(minus_log_posterior, points)
    else:
        results = list(map(minus_log_posterior, points))

    post=results[0]
    grad=np.array([(results[i+1]-post)/maxlike_sampler.epsilon  for i in range(n)])
    print("Norm of gradient: {:.2e}".format(np.linalg.norm(grad)))
    return post, grad




class PmaxlikeSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("like", float), ("post", float)]

    def config(self):
        self.method = self.read_ini("method", str, "CG")
        self.tolerance = self.read_ini("tolerance", float, 1e-3)
        self.maxiter = self.read_ini("maxiter", int, 1000)
        self.output_ini = self.read_ini("output_ini", str, "")
        self.output_cov = self.read_ini("output_covmat", str, "")
        self.epsilon = self.read_ini("gradient_epsilon",float,1e-9)
        self.gradient_tolerance = self.read_ini("gradient_tolerance",float,1e-5)
        self.converged = False
        global maxlike_sampler
        maxlike_sampler = self

    def execute(self):
        import scipy.optimize


        #starting position in the normalized space
        start_vector = self.pipeline.normalize_vector(self.pipeline.start_vector())
        bounds = [(0.0, 1.0) for p in self.pipeline.varied_params]


        result = scipy.optimize.minimize(posterior_and_gradient, start_vector, method=self.method,
          jac=True, tol=self.tolerance,  #bounds=bounds, 
          options={'maxiter':self.maxiter, 'disp':True, 'eps':self.epsilon, 'gtol':self.gradient_tolerance})

        opt_norm = result.x
        opt = self.pipeline.denormalize_vector(opt_norm)
        
        results = self.pipeline.run_results(opt)

        #Some output - first log the parameters to the screen.
        print("Best fit:\n%s"%'   '.join(str(x) for x in opt))
        print("Best likelihood: %f\n", results.like)

        #Next save them to the proper table file
        self.output.parameters(opt, results.extra, results.prior, results.like, results.post)

        #If requested, create a new ini file for the
        #best fit.
        if self.output_ini:
          self.pipeline.create_ini(opt, self.output_ini)

        # If we are coupling to later samplers they can use the peak we have found.
        self.distribution_hints.set_peak(opt, results.post)

        #Also if requested, approximate the covariance matrix with the 
        #inverse of the Hessian matrix.
        #For a gaussian likelihood this is exact.
        covmat = None
        if hasattr(result, 'hess_inv'):
            covmat = self.pipeline.denormalize_matrix(result.hess_inv)
        elif hasattr(result, 'hess'):
            covmat = self.pipeline.denormalize_matrix(np.linalg.inv(result.hess_inv))

        if covmat is None:
            if self.output_cov:
               print("Sorry - the optimization method you chose does not return a covariance (or Hessian) matrix")
        else:
            if self.output_cov:
                np.savetxt(self.output_cov, covmat)
            self.distribution_hints.set_cov(covmat)

        self.converged = True

    def is_converged(self):
        return self.converged

# legacy alias
PMaxlikeSampler = PmaxlikeSampler



from . import lazy_pylab as pylab

class MiniTable:
    """
    A simple class for storing and writing out a table of data.

    This is to avoid imposing a dependency on astropy tables.

    """
    def __init__(self, cols):
        self.cols = cols
        self.rows = []

    def to_astropy(self):
        from astropy.table import Table
        return Table(rows=self.rows, names=self.cols)

    def append(self, row):
        if len(row)!=len(self.cols):
            raise ValueError("Row has wrong number of columns")
        self.rows.append(row)

    def write(self, filename):
        with open(filename, "w") as f:
            f.write("#")
            f.write(" ".join(self.cols))
            f.write("\n")
            for row in self.rows:
                f.write("  ".join(str(x) for x in row))
                f.write("\n")
    
    def __len__(self):
        return len(self.rows)

    def __getitem__(self, row_or_col):
        if isinstance(row_or_col, int):
            return self.rows[row_or_col]
        else:
            i = self.cols.index(row_or_col)
            return np.array([row[i] for row in self.rows])


class PostprocessProduct(object):
    def __init__(self, name, filename, value, info=None):
        self.name = name
        self.filename = filename
        self.value = value
        self.info = info

    def save(self):
        pass

class PostprocessPlot(PostprocessProduct):
    def save(self):
        pylab.figure(self.value.number)
        pylab.savefig(self.filename)
        pylab.close()

    def tweak(self, tweak):
        print("Tweaking", self.name)
        pylab.figure(self.value.number)
        tweak.info = self.info
        tweak.run()

class PostprocessTable(PostprocessProduct):
    def save(self):
        self.value.write(self.filename)

class PostprocessText(PostprocessProduct):
    def save(self):
        self.value.seek(0)
        text = self.value.read()
        with open(self.filename, "w") as f:
            f.write(text)
        self.value.close()


import ctypes as ct
from . import dbt_types
import os

source_dir = os.path.dirname(__file__)
libcosmosis_path = os.path.join(source_dir, os.path.pardir, "libcosmosis.so")
dll = ct.cdll.LoadLibrary(libcosmosis_path)

# We export a symbol in the C code to tell us this
enum_size = ct.c_int.in_dll(dll, "cosmosis_enum_size").value

#Assuming that enums are signed ...
c_enum = {
	1:ct.c_int8,
	2:ct.c_int16,
	4:ct.c_int32,
	8:ct.c_int64,
}[enum_size]

free = dll.free
free.argtypes = [ct.c_void_p]
free.restype = None

c_block = ct.c_size_t
c_datatype = c_enum
c_status = ct.c_int
c_str = ct.c_char_p
c_int = ct.c_int
c_int_p = ct.POINTER(ct.c_int)


def load_library_function(namespace, name, argtypes, restype):
	function = getattr(dll,name)
	function.argtypes = argtypes
	function.restype = restype
	namespace[name] = function

class c_complex(ct.Structure):
	_fields_ = [("real", ct.c_double), ("imag", ct.c_double)]


def load_function_types(namespace, c_type, c_name):
	load_library_function(namespace, "c_datablock_put_%s"%c_name, [c_block, c_str, c_str, c_type], c_status)
	load_library_function(namespace, "c_datablock_replace_%s"%c_name, [c_block, c_str, c_str, c_type], c_status)
	load_library_function(namespace, "c_datablock_get_%s"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type)], c_status)
	load_library_function(namespace, "c_datablock_get_%s_default"%c_name, [c_block, c_str, c_str, c_type, ct.POINTER(c_type)], c_status)

def load_array_function_types(namespace, c_type, c_name):
	load_library_function(namespace, "c_datablock_put_%s_array_1d"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int], c_status)
	load_library_function(namespace, "c_datablock_replace_%s_array_1d"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int], c_status)
	load_library_function(namespace, "c_datablock_get_%s_array_1d"%c_name, [c_block, c_str, c_str, ct.POINTER(ct.POINTER(c_type)), c_int_p], c_status)
	load_library_function(namespace, "c_datablock_get_%s_array_shape"%c_name, [c_block, c_str, c_str, c_int, c_int_p], c_status)
	load_library_function(namespace, "c_datablock_get_%s_array"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int, c_int_p], c_status)
	load_library_function(namespace, "c_datablock_put_%s_array"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int, c_int_p], c_status)
	load_library_function(namespace, "c_datablock_replace_%s_array"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int, c_int_p], c_status)
	load_library_function(namespace, "c_datablock_get_%s_array_1d_preallocated"%c_name, [c_block, c_str, c_str, ct.POINTER(c_type), c_int_p, c_int], c_status)

load_function_types(locals(), ct.c_int, 'int')
load_function_types(locals(), ct.c_bool, 'bool')
load_function_types(locals(), ct.c_double, 'double')
load_function_types(locals(), c_complex, 'complex')
load_function_types(locals(), c_str, 'string')

load_array_function_types(locals(), ct.c_int, 'int')
load_array_function_types(locals(), ct.c_double, 'double')
#load_array_function_types(locals(), ct.c_complex, 'complex')

load_library_function(
	locals(), 
	"make_c_datablock",
	[],
	c_block
)

load_library_function(
       locals(), 
       "destroy_c_datablock",
       [c_block],
       c_int
)

load_library_function(
	locals(), 
	"clone_c_datablock",
	[c_block],
	c_block
)



load_library_function(
	locals(), 
	"c_datablock_num_sections",
	[c_block],
	c_int
)

load_library_function(
	locals(),
	"c_datablock_get_array_length",
	[c_block, c_str, c_str],
	c_int
	)

load_library_function(
	locals(),
	"c_datablock_get_type",
	[c_block, c_str, c_str, ct.POINTER(c_datatype)],
	c_status
	)

load_library_function(
	locals(),
	"c_datablock_has_section",
	[c_block, c_str],
	ct.c_bool
	)

load_library_function(
	locals(),
	"c_datablock_has_value",
	[c_block, c_str, c_str],
	ct.c_bool
	)

load_library_function(
	locals(),
	"c_datablock_get_section_name",
	[c_block, ct.c_int],
	c_str
	)

load_library_function(
	locals(),
	"c_datablock_get_value_name",
	[c_block, c_str, ct.c_int],
	c_str
	)

load_library_function(
	locals(),
	"c_datablock_get_value_name_by_section_index",
	[c_block, ct.c_int, ct.c_int],
	c_str
	)

load_library_function(
	locals(),
	"c_datablock_num_values",
	[c_block, c_str],
	ct.c_int
	)


load_library_function(
	locals(),
	"c_datablock_delete_section",
	[c_block, c_str],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_copy_section",
	[c_block, c_str, c_str],
	ct.c_int
	)


load_library_function(
	locals(),
	"c_datablock_report_failures",
	[c_block],
	ct.c_int
	)


load_library_function(
	locals(),
	"c_datablock_print_log",
	[c_block],
	ct.c_int
	)


load_library_function(
	locals(),
	"c_datablock_log_access",
	[c_block, c_str, c_str, c_str],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_get_log_count",
	[c_block],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_get_log_entry",
	[c_block, ct.c_int, ct.c_int, c_str, c_str, c_str, c_str],
	c_status
	)




load_library_function(
	locals(),
	"c_datablock_put_metadata",
	[c_block, c_str, c_str, c_str, c_str],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_replace_metadata",
	[c_block, c_str, c_str, c_str, c_str],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_get_metadata",
	[c_block, c_str, c_str, c_str, ct.POINTER(c_str)],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_put_double_grid",
	[c_block, c_str, 
	    c_str, ct.c_int, ct.POINTER(ct.c_double),
	    c_str, ct.c_int, ct.POINTER(ct.c_double),
	    c_str, ct.POINTER(ct.POINTER(ct.c_double)),
	],
	ct.c_int
	)
 

load_library_function(
	locals(),
	"c_datablock_replace_double_grid",
	[c_block, c_str, 
	    c_str, ct.c_int, ct.POINTER(ct.c_double),
	    c_str, ct.c_int, ct.POINTER(ct.c_double),
	    c_str, ct.POINTER(ct.POINTER(ct.c_double)),
	],
	ct.c_int
	)


load_library_function(
	locals(),
	"c_datablock_get_array_ndim",
	[c_block, c_str, c_str, c_int_p],
	ct.c_int
	)

load_library_function(
	locals(),
	"c_datablock_put_str_array_1d",
	[c_block, c_str, c_str, ct.POINTER(c_str), ct.c_int],
	ct.c_int
)


# Python will auto-convert returned char pointers to strings,
# but that will leak memory here, so use a void pointer instead
load_library_function(
	locals(),
	"c_datablock_get_str_array_1d_preallocated",
	[c_block, c_str, c_str, ct.POINTER(ct.c_void_p), c_int_p],
	ct.c_int
)

load_library_function(
	locals(),
	"enable_cosmosis_segfault_handler",
	[],
	None
)



from .elements import PostProcessorElement
from cosmosis.runtime.config import Inifile
import os
import tempfile
import sys
from collections import defaultdict

def ini_from_header(header_text):
	"""
	Parse a cosmosis-format set of header comments that encode the
	parameters and values used in a run.

	These lines use the sentinels:
	START_OF_PARAMS_INI
	END_OF_PARAMS_INI
	START_OF_VALUES_INI
	END_OF_VALUES_INI
	to mark sections.

	"""
	state=None
	param_lines = []
	value_lines = []
	prior_lines = []
	for line in header_text:
		line=line.strip()
		if line=='START_OF_PARAMS_INI':
			state='params'
			continue
		elif line=='END_OF_PARAMS_INI':
			state=None
			continue
		elif line=='START_OF_VALUES_INI':
			state='values'
			continue
		elif line=='END_OF_VALUES_INI':
			state=None
			continue
		elif line=='START_OF_PRIORS_INI':
			state='priors'
			continue
		elif line=='END_OF_PRIORS_INI':
			state=None
			continue
		if state=='params':
			param_lines.append(line+"\n")
		elif state=='values':
			value_lines.append(line+"\n")
		elif state=='priors':
			prior_lines.append(line+"\n")

	f = Inifile(None)
	f.read_string(str("\n".join(param_lines)))
	g = Inifile(None)
	g.read_string(str("\n".join(value_lines)))
	h = Inifile(None)
	h.read_string(str("\n".join(prior_lines)))

	return f, g, h




#


class Rerunner(PostProcessorElement):
	def __init__(self, dirname, *args, **kwargs):
		self.rerun_dirname=dirname
		super(Rerunner, self).__init__(*args, **kwargs)

	def test_run_sample(self, sample):

		#Turn the output header into an ini file.
		#Definitely better to do it here as any
		#envar substitution is already done
		params_ini, values_ini, priors_ini = ini_from_header(self.source.comments[0])
		params_file = tempfile.NamedTemporaryFile(suffix='.ini', mode='w+')
		values_file = tempfile.NamedTemporaryFile(suffix='.ini', mode='w+')
		priors_file = tempfile.NamedTemporaryFile(suffix='.ini', mode='w+')

		#Now override the old parameters so that we:
		# - run the test sampler not the old sampelr
		# - save to our temp dir
		# - read from our temp files

		try:
			params_ini.add_section('test')
		except:
			pass

		likes = params_ini.get('pipeline', 'likelihoods', fallback='')
		params_ini.set('pipeline', 'values', values_file.name)
		params_ini.set('pipeline', 'priors', priors_file.name)
		params_ini.set('pipeline', 'timing', 'T')
		params_ini.set('pipeline', 'debug', 'T')
		params_ini.set('runtime', 'sampler', 'test')
		params_ini.set('test', 'save_dir', self.rerun_dirname)


		params_ini.write(params_file)
		params_file.flush()

		nvaried = self.source.metadata[0]['n_varied']
		varied_params = self.source.colnames[:nvaried]
		values_update = []
		i=0
		print("Updating these parameters:")
		for (sec,key), val in values_ini:
			if '{}--{}'.format(sec,key) in varied_params:
				l, _, u = val.split()
				print("    {}--{} = {}".format(sec,key,sample[i]))
				value_string = "{}  {}  {}".format(l, sample[i], u)
				values_update.append((sec, key, value_string))
				i+=1
			else:
				values_update.append((sec, key, val))
		print("")
		for sec, key, val in values_update:
			values_ini.set(sec, key, val)

		values_ini.write(values_file)
		values_file.flush()


		# No changes needed to prior
		priors_ini.write(priors_file)
		priors_file.flush()

		#A bit of painful fiddling to work around
		#link issues on a mac.  To run the postprocess
		#command we had to drop a problematic entry
		#in DYLD_LIBRARY_PATH.  But to run cosmosis
		#we need it back!
		if "UPS_DIR" in os.environ and os.uname()[0]=='Darwin':
			extra_ld = '{0}/lib'.format(os.environ['GCC_FQ_DIR'])
			old_ld = os.environ['DYLD_LIBRARY_PATH']
			new_ld = 'DYLD_LIBRARY_PATH="{0}:{1}"'.format(old_ld, extra_ld)
		else:
			new_ld = ''

		#Run the actual command
		cmd = "{0} cosmosis {1}".format(new_ld, params_file.name)
		status = os.system(cmd)

		#Now we have the cosmology data!
		return status




class BestFitRerunner(Rerunner):
	"Re-run sample(s) from an existing chain under the test sampler"

	def run(self):
		best_fit_index = self.source.get_col("post").argmax()
		sample = self.source.get_row(best_fit_index)
		print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
		print("Re-running maximum-posterior sample and saving results to {}".format(self.rerun_dirname))
		print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
		self.test_run_sample(sample)


		return []



from cosmosis.runtime import parameter
from cosmosis import run_cosmosis, Inifile
from cosmosis.output import InMemoryOutput
import numpy as np
import tempfile
import os

def test_override_new():
    with tempfile.NamedTemporaryFile('w') as values:
        values.write(
            "[parameters]\n"
            "p1=-3.0  0.0  3.0\n"
            "p2=-3.0  0.0  3.0\n"
            "p3=20.0\n")
        values.flush()
        print(values.name)
        params = parameter.Parameter.load_parameters(values.name, override=None)

    assert params[0] == ("parameters", "p1")
    assert params[0].start == 0.0
    assert np.allclose(params[0].limits, [-3, 3])

    assert params[1] == ("parameters", "p2")
    assert params[1].start == 0.0

    assert params[2] == ("parameters", "p3")
    assert np.isclose(params[2].start, 20.0)


    with tempfile.NamedTemporaryFile('w') as values:
        values.write(
            "[parameters]\n"
            "p1=-3.0  0.0  3.0\n"
            "p2=-3.0  0.0  3.0\n"
            "p3=20.0\n")
        values.flush()
        override = {
            ("parameters", "p1"): "1.0",
            ("xxx", "aaa"): "100.0",
        }
        params = parameter.Parameter.load_parameters(values.name, override=override)
    

    assert params[0] == ("parameters", "p1")
    assert params[0].start == 1.0
    assert np.allclose(params[0].limits, 1.0)

    assert params[1] == ("parameters", "p2")
    assert params[1].start == 0.0
    assert np.allclose(params[1].limits, [-3, 3])

    assert params[2] == ("parameters", "p3")
    assert np.isclose(params[2].start, 20.0)

    print(params)
    assert params[3] == ("xxx", "aaa")
    assert np.isclose(params[3].start, 100.0)
    assert np.allclose(params[3].limits, 100.0)


def test_header_update():
    root = os.path.split(os.path.abspath(__file__))[0]

    # Create a values file
    with tempfile.NamedTemporaryFile('w') as values:
        values.write(
            "[parameters]\n"
            "p1=-3.0  0.0  3.0\n"
            "p2=-3.0  0.0  3.0\n"
            "p3=20.0\n")
        values.flush()

        override = {
            ('runtime', 'root'): root,
            ('runtime', 'sampler'): 'emcee',
            ("pipeline", "debug"): "F",
            ("pipeline", "modules"): "test1",
            ("pipeline", "values"): values.name,
            ("test1", "file"): "example_module.py",
            ("emcee", "walkers"): "8",
            ("emcee", "samples"): "10"
        }
        ini = Inifile(None, override=override)
        output = InMemoryOutput()
        override_values = {
            ("parameters", "p1"): "-2.0  0.0  2.0",
        }

        run_cosmosis(ini, variables=override_values, output=output)
        print(output.comments)

        assert 'p1 = -2.0  0.0  2.0' in output.comments



#!/usr/bin/env python


import sys
import configparser

import argparse
import os
import pdb
import cProfile
from .runtime.config import Inifile, CosmosisConfigurationError
from .runtime.pipeline import LikelihoodPipeline
from .runtime import mpi_pool
from .runtime import logs
from .runtime import process_pool
from .runtime.utils import ParseExtraParameters, stdout_redirected, import_by_path, under_over_line, underline, overline
from .samplers.sampler import Sampler, ParallelSampler, Hints
from . import output as output_module
from .runtime.handler import activate_segfault_handling
from .version import __version__


RUNTIME_INI_SECTION = "runtime"


def demo_1_special (inifile):
    if "demo1.ini" in inifile:
        print("""
Congratulations: you have just run cosmosis demo one!

You can make plots of the outputs of this using this command:
    cosmosis-postprocess demos/demo1.ini -o outputs/demo1

Then you can try out the other demos...
... and read the information about plotting their output and what they are doing online.
Please get in touch with any problems, ideally by filing an Issue. Thanks!
""")


def demo_10_special(inifile):
    if   "demo10.ini" in inifile   and   not os.getenv ("HALOFIT", ""):
        print()
        print("Welcome to demo10!")
        print()
        print("**PLEASE NOTE**:")
        print()
        print("There are two flavours of this demo, selected through an ")
        print("environment variable called `HALOFIT'; this variable is not ")
        print("currently set, so we are giving it the value `takahashi'.")

        os.environ ["HALOFIT"] = "takahashi"



def demo_20a_special(inifile):
    if  "demo20a.ini" in inifile:
        print ()
        print ("You have completed demo20a, now run demo20b and compare")
        print ("results with demo5!")


    if   "demo20b.ini" in inifile   and   not os.path.isfile ("./demo20a.txt"):
        print(under_over_line("YOU MUST RUN demo20a BEFORE YOU CAN RUN demo20b.", chr='*'))

def sampler_main_loop(sampler, output, pool, is_root):
    # Run the sampler until convergence
    # which really means "finished" here - 
    # a sampler can "converge" just by reaching the 
    # limit of the number of samples it is allowed.
    if is_root:
        while not sampler.is_converged():
            sampler.execute()
            #Flush any output. This is to stop
            #a problem in some MPI cases where loads
            #of output is built up before being written.
            if output:
                output.flush()
        # If we are in parallel tell the other processors to end the 
        # loop and prepare for the next sampler
        if pool and sampler.is_parallel_sampler:
            pool.close()
    else:
        if sampler.is_parallel_sampler:
            sampler.worker()



def write_header_output(output, params, values, pipeline, values_override=None):
    # If there is an output file, save the ini information to
    # it as well.  We do it here since it's nicer to have it
    # after the sampler options that get written in sampler.config.
    # Create a buffer to store the output:
    output.comment("START_OF_PARAMS_INI")
    comment_wrapper = output.comment_file_wrapper()
    params.write(comment_wrapper)
    output.comment("END_OF_PARAMS_INI")
    # Do the same with the values file.
    # Unfortunately that means reading it in again;
    # if we ever refactor this bit we could eliminate that.
    if isinstance(values, Inifile):
        values_ini = values
    elif values is None:
        values_ini=Inifile(pipeline.values_filename, override=values_override)
    else:
        values_ini=values
    output.comment("START_OF_VALUES_INI")
    values_ini.write(comment_wrapper)
    output.comment("END_OF_VALUES_INI")

    # And the same with the priors
    output.comment("START_OF_PRIORS_INI")
    for priors_file in pipeline.priors_files:
        if isinstance(priors_file, Inifile):
            prior_ini = priors_file
        else:
            prior_ini=Inifile(priors_file)
        prior_ini.write(comment_wrapper)
    output.comment("END_OF_PRIORS_INI")

def setup_output(sampler_class, sampler_number, ini, pool, sample_methods, sample_method, resume, output):

    output_original = output

    needs_output = sampler_class.needs_output and \
       (pool is None or pool.is_master() or sampler_class.parallel_output)

    if not needs_output:
        return None

    if output_original is None:
        #create the output files and methods.
        try:
            output_options = dict(ini.items('output'))
        except configparser.NoSectionError:
            raise ValueError("ERROR:\nFor the sampler (%s) you chose in the [runtime] section of the ini file I also need an [output] section describing how to save results\n\n"%sample_method)
        #Additionally we tell the output here if
        #we are parallel or not.
        if (pool is not None) and (sampler_class.parallel_output):
            output_options['rank'] = pool.rank
            output_options['parallel'] = pool.size

        number_samplers = len(sample_methods)

        #Give different output filenames to the different sampling steps
        #Only change if this is not the last sampling step - the final
        #one retains the name in the output file.
        # Change, e.g. demo17.txt to demo17.fisher.txt
        if ("filename" in output_options) and (sampler_number<number_samplers-1):

            sampler_count = sample_methods.count(sample_method)
            filename = output_options['filename']
            filename, ext = os.path.splitext(filename)
            if sampler_count == 1:
                filename += '.' + sampler_class.name
                filename += ext
            else:
                sampler_repeat_index = 0
                for i in range(sampler_number):
                    if sample_methods[i] == sample_method:
                        sampler_repeat_index += 1
                filename += '.' + sampler_class.name + '.' + str(sampler_repeat_index)
                filename += ext
            output_options['filename'] = filename


        if ("filename" in output_options):
            print("* Saving output -> {}".format(output_options['filename']))

        #Generate the output from a factory
        output = output_module.output_from_options(output_options, resume)
    elif isinstance(output_original, output_module.OutputBase):
        output = output_original
    elif isinstance(output_original, str):
        if output_original == "astropy":
            output = output_module.AstropyOutput()
        elif output_original == "none":
            output = output_module.NullOutput()
        elif output_original == "in_memory":
            output = output_module.InMemoryOutput()
        else:
            raise ValueError(f"Unknown output option {output_original}")
    else:
        raise ValueError(f"Unknown output type {type(output_original)}")
        

    output.metadata("sampler", sample_method)

    return output


def run_cosmosis(ini, pool=None, pipeline=None, values=None, priors=None, override=None,
                 profile_mem=0, profile_cpu="", variables=None, only=None, output=None):
    """
    Execute cosmosis.

    Parameters
    ----------
    ini: str, cosmosis.Inifile, or None
        The parameter file from which to build the cosmosis run. If set to a string the
        file is read from disc. If set to None, the other parameters must contain
        all the required CosmoSIS parameters.
    
    pool: None, cosmosis.MPIPool, or cosmosis.process_pool.Pool
        A pool object to enable multi-process parallel execution. If left as the default
        None then the code is run with a single process (though modules may still run
        using OpenMP parallelism).
    
    pipeline: None or cosmosis.LikelihoodPipeline
        If set, ignore the pipeline definition in the ini file and use this pipeline
        instead.
    
    values: None or dict[str, str]->str
        If set, ignore the numerical parameter values in the ini file and use these 
        instead.
    
    priors: None or dict[str, str]->str
        If set, ignore the prior values in the ini file and use these 
        instead. 

    override: None or dict[str, str]->str
        If set, override parameter values in the ini file from the dictionary.
    
    profile_mem: int
        If changed from the default zero value, print a memory profile every 
        profile_mem seconds.

    profile_cpu: str
        If changed from the default empty string, print CPU profile information
        and also save to the named file. If running in parallel, save to 
        {profile_cpu}.{rank}.
    
    variables: None or dict[str, str]->str
        If set, override variable values in the ini file from the dictionary.

    only: None or str
        If set, fix all the variable values except the one supplied.

    output: None or cosmosis.Output
        If set, use this output object to save the results. If not set, create
        an output object from the ini file.
    """
    no_subprocesses = os.environ.get("COSMOSIS_NO_SUBPROCESS", "") not in ["", "0"]

    smp = isinstance(pool, process_pool.Pool)

    # Load configuration.
    is_root = (pool is None) or pool.is_master()
    ini_is_str = isinstance(ini, str)
    ini_original = ini
    output_original = output
    ini = Inifile(ini, override=override, print_include_messages=is_root)

    pre_script = ini.get(RUNTIME_INI_SECTION, "pre_script", fallback="")
    post_script = ini.get(RUNTIME_INI_SECTION, "post_script", fallback="")

    verbosity = ini.get(RUNTIME_INI_SECTION, "verbosity", fallback="")
    
    if not verbosity:
        if ini.has_option("pipeline", "quiet"):
            quiet = ini.getboolean("pipeline", "quiet", fallback=False)
            verbosity = "quiet" if quiet else "standard"
        else:
            verbosity = ini.get(RUNTIME_INI_SECTION, "output", fallback="standard")
    logs.set_verbosity(verbosity)

    if ini.has_option("pipeline", "quiet") and is_root:
        logs.warning("Deprecated: The [pipeline] quiet option is deprecated.  Set [runtime] verbosity instead.")

    if is_root and pre_script:
        if no_subprocesses:
            print("Warning: subprocesses not allowed on this system as")
            print("COSMOSIS_NO_SUBPROCESS variable was set.")
            print("Ignoring pre-script.")
        else:
            status = os.WEXITSTATUS(os.system(pre_script))
            if status:
                raise RuntimeError("The pre-run script {} retuned non-zero status {}".format(
                    pre_script, status))

    if is_root and profile_mem:
        from cosmosis.runtime.memmon import MemoryMonitor
        # This launches a memory monitor that prints out (from a new thread)
        # the memory usage every profile_mem seconds
        mem = MemoryMonitor.start_in_thread(interval=profile_mem)

    if profile_cpu:
        profile = cProfile.Profile()
        profile.enable()

    # Create pipeline.
    if pipeline is None:
        cleanup_pipeline = True
        pool_stdout = ini.getboolean(RUNTIME_INI_SECTION, "pool_stdout", fallback=False)
        if is_root:
            if ini_is_str:
                print(underline(f"Setting up pipeline from parameter file {ini_original}"))
            else:
                print(underline(f"Setting up pipeline from pre-constructed configuration"))

        if is_root or pool_stdout:
            pipeline = LikelihoodPipeline(ini, override=variables, values=values, only=only, priors=priors)
        else:
            if pool_stdout:
                pipeline = LikelihoodPipeline(ini, override=variables, values=values, only=only, priors=priors)
            else:
                # Suppress output on everything except the root process
                with stdout_redirected():
                    pipeline = LikelihoodPipeline(ini, override=variables, values=values, only=only, priors=priors)

        if pipeline.do_fast_slow:
            pipeline.setup_fast_subspaces()
    else:
        # We should not cleanup a pipeline which we didn't make
        cleanup_pipeline = False

    # This feature lets us import additional samplers at runtime
    sampler_files = ini.get(RUNTIME_INI_SECTION, "import_samplers", fallback="").split()
    for i, sampler_file in enumerate(sampler_files):
        # give the module a new name to avoid name clashes if people
        # just call their thing by the same name
        import_by_path('additional_samplers_{}'.format(i), sampler_file)



    # determine the type(s) of sampling we want.
    sample_methods = ini.get(RUNTIME_INI_SECTION, "sampler", fallback="test").split()

    for sample_method in sample_methods:
        if sample_method not in Sampler.registry:
            raise ValueError("Unknown sampler method %s" % (sample_method,))

    #Get that sampler from the system.
    sampler_classes = [Sampler.registry[sample_method] for sample_method in sample_methods]

    if pool:
        if not any(issubclass(sampler_class,ParallelSampler) for sampler_class in sampler_classes):
            if len(sampler_classes)>1:
                raise ValueError("None of the samplers you chose support parallel execution!")
            else:
                raise ValueError("The sampler you chose does not support parallel execution!")
        for sampler_class in sampler_classes:
            if isinstance(pool, process_pool.Pool) and issubclass(sampler_class,ParallelSampler) and not sampler_class.supports_smp:
                name = sampler_class.__name__[:-len("Sampler")].lower()
                raise ValueError("Sorry, the {} sampler does not support the --smp flag.".format(name))

    number_samplers = len(sampler_classes)

    if ini.has_option("pipeline", "failure_log"):
        filename = ini.get("pipeline", "failure_log")
        failure_log_file = mpi_pool.MPILogFile(filename, pool)
        pipeline.set_failure_log_file(failure_log_file)
    else:
        failure_log_file = None



    #To start with we do not have any estimates of 
    #anything the samplers might give us like centers
    #or covariances. 
    distribution_hints = Hints()

    #Now that we have a sampler we know whether we will need an
    #output file or not.  By default new samplers do need one.
    for sampler_number, (sampler_class, sample_method) in enumerate(
            zip(sampler_classes, sample_methods[:])):
        sampler_name = sampler_class.__name__[:-len("Sampler")].lower()

        # The resume feature lets us restart from an existing file.
        # It's not fully rolled out to all the suitable samplers yet though.
        resume = ini.getboolean(RUNTIME_INI_SECTION, "resume", fallback=False)

        # Polychord, multinest, and nautilus have their own internal
        # mechanism for resuming chains.
        if sampler_class.internal_resume:
            resume2 = ini.getboolean(sampler_name, "resume", fallback=False)
            resume = resume or resume2

            if resume and is_root:
                print(f"Resuming sampling using {sampler_name} internal mechanism, "
                      "so starting a new output chain file.")

            # Tell the sampler to resume directly
            if not ini.has_section(sampler_name):
                ini.add_section(sampler_name)
            ini.set(sampler_name, "resume", str(resume))

            # Switch off the main cosmosis resume mechanism
            resume = False

        # Not all samplers can be resumed.
        if resume and not sampler_class.supports_resume:
            print("NOTE: You set resume=T in the [runtime] section but the sampler {} does not support resuming yet.  I will ignore this option.".format(sampler_name))
            resume=False
        

        if is_root:
            print("****************************************************")
            print("* Running sampler {}/{}: {}".format(sampler_number+1,number_samplers, sampler_name))
            if pool and smp:
                print(f"* Using multiprocessing (SMP) with {pool.size} processes.")
            elif pool:
                print(f"* Using MPI with {pool.size} processes.")
            else:
                print("* Running in serial mode.")

        output = setup_output(sampler_class, sampler_number, ini, pool, sample_methods, sample_method, resume, output_original)

        if is_root:
            print("****************************************************")

        #Initialize our sampler, with the class we got above.
        #It needs an extra pool argument if it is a ParallelSampler.
        #All the parallel samplers can also act serially too.
        if pool and sampler_class.is_parallel_sampler:
            sampler = sampler_class(ini, pipeline, output, pool)
        else:
            sampler = sampler_class(ini, pipeline, output)
         
        #Set up the sampler - for example loading
        #any resources it needs or checking the ini file
        #for additional parameters.
        sampler.distribution_hints.update(distribution_hints)
        sampler.config()

        # Potentially resume
        if resume and sampler_class.needs_output and \
            sampler_class.supports_resume and \
            (is_root or sampler_class.parallel_output):
           sampler.resume()

        if output:
            write_header_output(output, ini, values, pipeline, values_override=variables)

        sys.stdout.flush()
        sys.stderr.flush()


        sampler_main_loop(sampler, output, pool, is_root)
        distribution_hints.update(sampler.distribution_hints)

        # get total number of evaluations on all MPI processes
        if pool is None:
            run_count_total = pipeline.run_count
            run_count_ok_total = pipeline.run_count_ok
        else:
            run_count_total = pool.allreduce(pipeline.run_count)
            run_count_ok_total = pool.allreduce(pipeline.run_count_ok)
        
        if is_root and sampler_name != 'test':
            logs.overview(f"Total posterior evaluations = {run_count_total} across all processes")
            logs.overview(f"Successful posterior evaluations = {run_count_ok_total} across all processes")
            if output:
                output.final("evaluations", run_count_total)
                output.final("successes", run_count_ok_total)
                output.final("complete", "1")


        if output:
            output.close()

    if cleanup_pipeline:
        pipeline.cleanup()

    if failure_log_file is not None:
        failure_log_file.close()

    if profile_cpu:
        profile.disable()
        if (pool is not None) and (not smp):
            profile_name = profile_cpu + f'.{pool.rank}'
        else:
            profile_name = profile_cpu
        profile.dump_stats(profile_name)
        profile.print_stats("cumtime")

    if is_root and profile_mem:
        mem.stop()

    # Extra-special actions we take to help new users playing with the demos
    demo_1_special(str(ini))
    demo_10_special(str(ini))
    demo_20a_special (str(ini))

    # User can specify in the runtime section a post-run script to launch.
    # In general this may be less useful than the pre-run script, because
    # often chains time-out instead of actually completing.
    # But we still offer it
    if post_script and is_root:
        # This decodes the exist status
        if no_subprocesses:
            print("Warning: subprocesses not allowed on this system as")
            print("COSMOSIS_NO_SUBPROCESS variable was set.")
            print("Ignoring post-script.")
        else:
            status = os.WEXITSTATUS(os.system(post_script))
            if status:
                sys.stdout.write("WARNING: The post-run script {} failed with error {}".format(
                    post_script, status))

    if isinstance(output_original, str):
        if output_original == "astropy":
            return 0, output.table
        else:
            return 0, output
    else:
        return 0


def make_graph(inifile, dotfile, params=None, variables=None):
    """
    Make a graphviz "dot" format file, describing the pipeline
    and how data is passed from section to section.

    Requires pygraphviz.

    Parameters
    ----------

    inifile: str
        A path to a pipeline file or an Inifile object

    dotfile: str
        Path to the new graph output file

    params: dict or None
        Dictionary of parameter overrides

    variables: dict or None
        Dictionary of value overrides
    """
    ini = Inifile(inifile, override=params)
    pipeline = LikelihoodPipeline(ini, override=variables)
    data = pipeline.run_parameters(pipeline.start_vector())
    pipeline.make_graph(data, dotfile)


# Make this global because it is useful for testing
parser = argparse.ArgumentParser(description="Run a pipeline with a single set of parameters", add_help=True)
parser.add_argument("inifile", help="Input ini file of parameters")
parser.add_argument("--mpi",action='store_true',help="Run in MPI mode.")
parser.add_argument("--smp",type=int,default=0,help="Run with the given number of processes in shared memory multiprocessing (this is experimental and does not work for multinest).")
parser.add_argument("--pdb",action='store_true',help="Start the python debugger on an uncaught error. Only in serial mode.")
parser.add_argument("--segfaults", "--experimental-fault-handling", action='store_true',help="Activate a mode that gives more info on segfault")
parser.add_argument("--mem", type=int, default=0, help="Print out memory usage every this many seconds from root process")
parser.add_argument("-p", "--params", nargs="*", action=ParseExtraParameters, help="Override parameters in inifile, with format section.name1=value1 section.name2=value2...")
parser.add_argument("-v", "--variables", nargs="*", action=ParseExtraParameters, help="Override variables in values file, with format section.name1=value1 section.name2=value2...")
parser.add_argument("--only", nargs="*", help="Fix all parameters except the ones listed")
parser.add_argument("--graph", type=str, default='', help="Do not run a sampler; instead make a graphviz dot file of the pipeline")
parser.add_argument('--version', action='version', version=__version__, help="Print out a version number")
parser.add_argument('--profile' , help="Save profiling (timing) information to this named file")


def main():
    try:
        args = parser.parse_args(sys.argv[1:])

        if args.graph:
            make_graph(args.inifile, args.graph, args.params, args.variables)
            return 0

        if args.segfaults:
            activate_segfault_handling()

        # initialize parallel workers
        if args.mpi:
            with mpi_pool.MPIPool() as pool:
                return run_cosmosis(ini=args.inifile, pool=pool, override=args.params, profile_mem=args.mem, profile_cpu=args.profile, variables=args.variables, only=args.only)
        elif args.smp:
            with process_pool.Pool(args.smp) as pool:
                return run_cosmosis(ini=args.inifile, pool=pool, override=args.params, profile_mem=args.mem, profile_cpu=args.profile, variables=args.variables, only=args.only)
        else:
            try:
                return run_cosmosis(ini=args.inifile, pool=None, override=args.params, profile_mem=args.mem, profile_cpu=args.profile, variables=args.variables, only=args.only)
            except Exception as error:
                if args.pdb:
                    print("There was an exception - starting python debugger because you ran with --pdb")
                    print(error)
                    pdb.post_mortem()
                else:
                    raise
    except CosmosisConfigurationError as e:
        print(e)
        return 1


if __name__=="__main__":
    status = main()
    sys.exit(status)



#!/usr/bin/env python

import os
import argparse
import numpy as np
#Do these as absolute imports instead of relative as
#we may want to run this program as a script directly.
from cosmosis.postprocessing import lazy_pylab as pylab
from cosmosis.runtime import utils


#Set up the command line arguments
parser = argparse.ArgumentParser(description="""
This script makes some standard plots based on a directory of CosomoSIS output
for a single set of cosmological parameters, as generated by the "test" sampler
in CosmoSIS.  Have a look at the first demo we have for an example.

You need to supply at least the name of the directory where CosmoSIS data was
saved.

You can either call this script directly or use the "postprocess" program on the
ini file.
""")

parser.add_argument("dirname", help="A directory of cosmology data as produced by the 'test' sampler")
parser.add_argument("-o", "--output_dir", default=".", help="The directory in which to put the plots")
parser.add_argument("-p", "--prefix", default="", help="A filename prefix for all the plots")
parser.add_argument("-t", "--type", default="png", help="Image file type suffix")


# Okay, here is the structure of this script.  We have a base Plot class which
# handles finding the data files, working out filenames, and saving plots.  The
# subclasses are then in charge of the plot contents.

# The Plot class also uses a metaclass to register its subclasses
# so that there is less plumbing to do when adding a new plot.

class RegisteredPlot(type):
    registry = set()
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        meta.registry.add(cls)
        meta.registry -= set(bases)
        return cls

class Plot(metaclass=RegisteredPlot):
    #Subclasses should override this to specify the base
    #part of their filename
    filename = "error"

    def __init__(self, dirname, outdir, prefix, suffix,
             quiet=False, figure=None):
        #Set up the plotting figure
        if figure is None:
            self.figure = pylab.figure()
        else:
            self.figure = figure

        #Can do prefixes if we want to all the filenames
        if prefix:
            prefix += "_"

        # We have an output directory, a prefix, base filename (which is
        # overridden by subclasses) And a suffix (file type).
        self.outfile = (outdir + "/" + prefix + self.filename
                    + "." + suffix)

        #All the data will be in subclasses of this
        self.dirname = dirname
        self.quiet=quiet

    def file_path(self, section, name):
        return self.dirname + "/" + section + "/" + name + ".txt"

    def load_file(self, section, name):
        "Load a data file from the directory of saved theory data"
        #Find the filename and load it
        filename = self.file_path(section, name)
        try:
            return np.loadtxt(filename)
        except Exception as e:
            raise IOError("Not making plot: %s (no data in this sample)"% self.__class__.__name__[:-4])

    #Handy little method for trying to numeric-ify a value
    @staticmethod
    def try_numeric(value):
        try:
            return float(value)
        except:
            pass
        try:
            return int(value)
        except:
            pass
        return value

    # The scalar parameters in the DataBlock are saved to a
    # file called values.txt for each section.
    # This file loads that values file into a dictionary
    def load_values(self, section):
        filename = self.dirname + "/" + section + "/values.txt"
        values = {}
        for line in open(filename):
            line=line.strip()
            if not line:
                continue
            name, value = line.split('=')
            values[name.strip()] = self.try_numeric(value.strip())
        return values

    #Base method - subclasses should call this super method
    #and then override to supply their own plot content
    def plot(self):
        pylab.figure(self.figure.number)

    #Need not be over-ridden
    def save(self):
        pylab.figure(self.figure.number)
        if not self.quiet:
            print("Saving ", self.outfile)
        pylab.savefig(self.outfile)

    #Need not be overridden. Called by the main function
    @classmethod
    def make(cls, dirname, outdir, prefix, suffix):
        p = cls(dirname, outdir, prefix, suffix)
        p.plot()
        p.save()
        return p.filename



class DistancePlot(Plot):
    "Subclasses of this do distance plots as a function of z"
    scaling=1.0
    def plot(self):
        super(DistancePlot, self).plot()
        z = self.load_file("distances", "z")
        d = self.load_file("distances", self.distance)
        pylab.plot(z, d*self.scaling)
        pylab.grid()
        pylab.xlabel("Redshift z")
        pylab.ylabel(self.name)



class AngularDiameterDistancePlot(DistancePlot):
    distance = "d_a"
    filename = "angular_distance"
    name = "Angular Diameter Distance $D_A / Mpc$"

class LuminosityDistancePlot(DistancePlot):
    distance = "d_l"
    filename = "luminosity_distance"
    name = "Luminosity Distance $D_L / Mpc$"

class ComovingDistancePlot(DistancePlot):
    distance = "d_m"
    filename = "comoving_distance"
    name = "Comoving Distance $D_L / Mpc$"

class HubblePlot(DistancePlot):
    distance = "h"
    filename = "hubble"
    name = "Hubble Parameter $H(z) / (km/s/Mpc)$"
    scaling = 2.99792458e+05

class DistanceModulusPlot(DistancePlot):
    distance = "mu"
    filename = "distance_modulus"
    name = "Distance Modulus $\\mu$"

class CMBSpectrumPlot(Plot):
    "The four different C_ell plots are subclasses of this"
    def plot(self):
        super(CMBSpectrumPlot, self).plot()
        ell = self.load_file("cmb_cl", "ell")
        c_ell = self.load_file("cmb_cl", self.name)
        pylab.plot(ell, c_ell)
        pylab.grid()
        pylab.xlabel("$\\ell$")
        pylab.ylabel("$\\ell(\\ell+1) C_\\ell/2\\pi \\mathrm{"
                      + self.name.upper ()
                      + "} / uK^2$")

class TTPlot(CMBSpectrumPlot):
    name = filename = "tt"

class EEPlot(CMBSpectrumPlot):
    name = filename = "ee"

class TEPlot(CMBSpectrumPlot):
    name = filename = "te"

class BBPlot(CMBSpectrumPlot):
    name = filename = "bb"
    def plot(self):
        super(BBPlot,self).plot()
        pylab.xlim(0,400)



class GrandPlot(Plot):
    "Grand CMB plot with all four spectra"
    filename = "grand"
    def plot(self):
        super(GrandPlot, self).plot()
        ell = self.load_file("cmb_cl", "ell")
        for name in ['tt', 'ee', 'te', 'bb']:
            c_ell = self.load_file("cmb_cl", name)
            pylab.loglog(ell, abs(c_ell), label=name.upper())
        pylab.legend()
        pylab.grid()
        pylab.xlabel("$\\ell$")
        pylab.ylabel("$\\ell(\\ell+1) C_\\ell/2\\pi / uK^2$")



class MatterPowerPlot(Plot):
    "Matter power spectrum, maybe including non-linear plot too if available"
    filename = "matter_power"
    def plot_section(self, section, label, p_name='p_k'):
        kh = self.load_file(section, "k_h")
        z = self.load_file(section, "z")
        p = self.load_file(section, p_name)
        if (p<0).all(): p*=-1
        z = np.atleast_1d(z)
        kh = np.atleast_1d(kh)
        p = np.atleast_2d(p)
        nk = len(kh)
        nz = len(z)
        if p.shape==(nz,nk):
            p = p.T
        pylab.loglog(kh, p[:,0], label=label)

    def power_files_exist(self, name):
        exist = [os.path.exists(
                os.path.join(self.dirname, name, filename+".txt"))
            for filename in ["k_h", "z", "p_k"]]
        return all(exist)


    def plot(self):
        super(MatterPowerPlot, self).plot()
        done_any=False
        if self.power_files_exist("matter_power_lin"):
            self.plot_section("matter_power_lin", "Linear")
            done_any=True
        if self.power_files_exist("matter_power_nl"):
            self.plot_section("matter_power_nl", "Non-Linear")
            done_any=True
        if self.power_files_exist("matter_power_gal"):
            self.plot_section("matter_power_gal", "Galaxy")
            done_any=True
        if self.power_files_exist("matter_power_no_bao"):
            self.plot_section("matter_power_no_bao", "No BAO")
            done_any=True
        if self.power_files_exist("intrinsic_alignment_parameters"):
            self.plot_section("intrinsic_alignment_parameters", "Intrinsic-intrinsic", p_name='p_ii')
            done_any=True
        if self.power_files_exist("intrinsic_alignment_parameters"):
            self.plot_section("intrinsic_alignment_parameters", "Shear-intrinsic", p_name='p_gi')
            done_any=True
        if not done_any:
            raise IOError("Not making plot: %s (no data in this sample)"% self.__class__.__name__[:-4])
        pylab.xlabel("$k / (Mpc/h)$")
        pylab.ylabel("$P(k) / (h^1 Mpc)^3$")
        pylab.grid()
        pylab.legend()



class ShearSpectrumPlot(Plot):
    "Shear-shear power spectrum"
    filename = "shear_power"
    ylim = (1e-8,1e-3)
    def plot(self):
        super(ShearSpectrumPlot, self).plot()
        self.plot_section("shear_cl")
        if os.path.exists(self.dirname + "/shear_cl_gg"):
            self.plot_section("shear_cl_gg")
        if os.path.exists(self.dirname + "/shear_cl_gi"):
            self.plot_section("shear_cl_gi")
        if os.path.exists(self.dirname + "/shear_cl_ii"):
            self.plot_section("shear_cl_ii")

    def plot_section(self, section):
        nbin = 0
        for i in range(1,100):
            filename = self.file_path(section,
                              "bin_{0}_{0}".format (i))
            if os.path.exists(filename):
                nbin += 1
            else:
                break
        if nbin==0:
            IOError("No data for plot: %s"% self.__class__.__name__[:-4])

        ell = self.load_file(section, "ell")
        sz = 1.0/(nbin+2)
        for i in range(1, nbin+1):
            for j in range(1, i+1):
                rect = (i*sz,j*sz,sz,sz)
                self.figure.add_axes(rect)
                #pylab.ploy()
                #pylab.subplot(nbin, nbin, (nbin*nbin)-nbin*(j-1)+i)
                cl = self.load_file(section,
                                "bin_{0}_{1}".format(i,j))

                if all(cl<=0):
                    cl *= -1
                pylab.loglog(ell, ell*(ell+1.) * cl/2/np.pi)
                pylab.ylim(*self.ylim)
                if i==1 and j==1:
                    pylab.xlabel("$\\ell$")
                    pylab.ylabel("$\\ell (\\ell+1) C_\\ell / 2 \\pi$")
                else:
                    pylab.gca().xaxis.set_ticklabels([])
                    pylab.gca().yaxis.set_ticklabels([])
                pylab.gca().tick_params(length=0.0, which='minor')
                pylab.gca().tick_params(length=3.0, which='major')
                pylab.gca().tick_params(labelsize=10)

                if section=="shear_cl":
                    pylab.text(1.5*ell.min(),
                               1.8e-4,
                               "(%d,%d)"%(i,j),
                                fontsize=8,
                                 color='red')
                    pylab.grid()


class MatterPower2D(ShearSpectrumPlot):
    "2D Matter power spectrum"
    filename = "matter_power_2d"
    ylim = (1e-6,1.0)
    def plot(self):
        super(ShearSpectrumPlot, self).plot()
        self.plot_section("matter_cl")


class ShearCorrelationPlusPlot(Plot):
    "Shear-shear correlation xi plus"
    filename = "shear_xi_plus"
    def plot(self):
        super(ShearCorrelationPlusPlot, self).plot()
        nbin = 0
        for i in range(1,100):
            filename = self.file_path("shear_xi_plus",
                                      "bin_{0}_{0}".format(i))
            if os.path.exists(filename):
                nbin += 1
            else:
                break
        if nbin==0:
            IOError("No data for plot: %s"% self.__class__.__name__[:-4])

        theta = self.load_file("shear_xi_plus", "theta")
        sz = 1.0/(nbin+2)
        for i in range(1, nbin+1):
            for j in range(1, i+1):
                rect = (i*sz,j*sz,sz,sz)
                self.figure.add_axes(rect)
                #pylab.ploy()
                #pylab.subplot(nbin, nbin, (nbin*nbin)-nbin*(j-1)+i)
                xi = self.load_file("shear_xi_plus",
                                "bin_{0}_{1}".format(i,j))
                pylab.loglog(theta, xi)
                pylab.xlim(1e-4,1e-1)
                pylab.ylim(2e-7,1e-3)
                if i==1 and j==1:
                    pylab.xlabel("$\\theta$")
                    pylab.ylabel("$\\xi_+(\\theta)$")
                else:
                    pylab.gca().xaxis.set_ticklabels([])
                    pylab.gca().yaxis.set_ticklabels([])

                pylab.gca().tick_params(length=0.0, which='minor')
                pylab.gca().tick_params(length=3.0, which='major')
                pylab.gca().tick_params(labelsize=10)

                pylab.text(1.5e-3,1.8e-4, "(%d,%d)"%(i,j), fontsize=8,
                           color='red')

                pylab.grid()



class ShearCorrelationMinusPlot(Plot):
    "Shear-shear correlation xi minus"
    filename = "shear_xi_minus"
    def plot(self):
        super(ShearCorrelationMinusPlot, self).plot()
        nbin = 0
        for i in range(1,100):
            filename = self.file_path("shear_xi_minus",
                                      "bin_{0}_{0}".format(i))
            if os.path.exists(filename):
                nbin += 1
            else:
                break
        if nbin==0:
            IOError("No data for plot: %s"% self.__class__.__name__[:-4])

        theta = self.load_file("shear_xi_minus", "theta")
        sz = 1.0/(nbin+2)
        for i in range(1, nbin+1):
            for j in range(1, i+1):
                rect = (i*sz,j*sz,sz,sz)
                self.figure.add_axes(rect)
                #pylab.ploy()
                #pylab.subplot(nbin, nbin, (nbin*nbin)-nbin*(j-1)+i)
                xi = self.load_file("shear_xi_minus",
                                "bin_{0}_{1}".format(i,j))
                pylab.loglog(theta, xi)
                pylab.xlim(1e-4,1e-1)
                pylab.ylim(2e-7,1e-3)
                if i==1 and j==1:
                    pylab.xlabel("$\\theta$")
                    pylab.ylabel("$\\xi_+(\\theta)$")
                else:
                    pylab.gca().xaxis.set_ticklabels([])
                    pylab.gca().yaxis.set_ticklabels([])

                pylab.gca().tick_params(length=0.0, which='minor')
                pylab.gca().tick_params(length=3.0, which='major')
                pylab.gca().tick_params(labelsize=10)

                pylab.text(1.5e-3,1.8e-4, "(%d,%d)"%(i,j), fontsize=8,
                           color='red')

                pylab.grid()


class GrowthPlot(Plot):
    filename='growth'
    def plot(self):
        super(GrowthPlot,self).plot()
        section = "growth_parameters"
        z = self.load_file(section, "z")
        d_z = self.load_file(section, "d_z")
        f_z = self.load_file(section, "f_z")
        pylab.plot(z, d_z, label='$d(z)$')
        pylab.plot(z, f_z, label='$f(z)$')
        pylab.grid()
        pylab.xlabel("Redshift z")
        pylab.ylabel("Growth Functions")
        pylab.legend(loc='center right')


class LuminositySlopePlot(Plot):
    filename='galaxy_luminosity_slope'
    def plot(self):
        super(LuminositySlopePlot,self).plot()
        section = "galaxy_luminosity_function"
        z = self.load_file(section, "z")
        alpha = self.load_file(section, "alpha")
        pylab.plot(z, alpha)
        pylab.grid()
        pylab.xlabel("Redshift z")
        pylab.ylabel("Luminosity Function Slope $\\alpha$")



def main(args):
    utils.mkdir(args.output_dir)
    for cls in Plot.registry:
        try:
            cls.make(args.dirname, args.output_dir, args.prefix, args.type)
        except IOError as err:
            print(err)


if __name__ == '__main__':
    main (parser.parse_args ())



from ..runtime.attribution import PipelineAttribution
from ..runtime.utils import get_git_revision
from ..runtime import Inifile
from ..output import InMemoryOutput
import datetime
import platform
import getpass
import os
import uuid
import pickle
from .hints import Hints
import numpy as np
import shutil
import numpy as np
import configparser

# Sampler metaclass that registers each of its subclasses

class RegisteredSampler(type):
    registry = {}
    def __new__(meta, name, bases, class_dict):
        if name.endswith("Sampler"):
            meta.registry = {name : cls for name, cls in meta.registry.items() if cls not in bases}
            config_name = name[:-len("Sampler")].lower()
            cls = type.__new__(meta, name, bases, class_dict)
            cls.name = config_name
            meta.registry[config_name] = cls
            return cls
        else:
            raise ValueError("Sampler classes must be named [Name]Sampler")

class Sampler(metaclass=RegisteredSampler):
    needs_output = True
    sampler_outputs = []
    understands_fast_subspaces = False
    parallel_output = False
    is_parallel_sampler = False
    supports_resume = False
    internal_resume = False

    
    def __init__(self, ini, pipeline, output=None):
        if isinstance(ini, Inifile):
            self.ini = ini
        else:
            self.ini = Inifile(ini)

        self.pipeline = pipeline
        # Default to an in-memory output
        if output is None:
            output = InMemoryOutput()
        self.output = output
        self.attribution = PipelineAttribution(self.pipeline.modules)
        self.distribution_hints = Hints()
        self.write_header()

    def write_header(self, output=None):
        if output is None:
            output = self.output
        if output is None:
            return

        for p in self.pipeline.output_names():
            output.add_column(p, float)
        for p,ptype in self.sampler_outputs:
            output.add_column(p, ptype)
        output.metadata("n_varied", len(self.pipeline.varied_params))
        self.attribution.write_output(output)
        for key, value in self.collect_run_metadata().items():
            output.metadata(key, value)
        blinding_header = self.ini.getboolean("output","blinding-header", fallback=False)
        if blinding_header:
            output.blinding_header()

    def collect_run_metadata(self):
        info = {
            'timestamp': datetime.datetime.now().isoformat(),
            'platform': platform.platform(),
            'platform_version': platform.version(),
            'uuid': uuid.uuid4().hex,
        }
        info['cosmosis_git_version'] = get_git_revision("$COSMOSIS_SRC_DIR")
        info['csl_git_version'] = get_git_revision("$COSMOSIS_SRC_DIR/cosmosis-standard-library")
        info['cwd_git_version'] = get_git_revision("$PWD")

        # The host name and username are (potentially) private information
        # so we only save those if privacy=False, which is not the default
        privacy = self.ini.getboolean('output','privacy', fallback=True)
        save_username = not privacy
        if save_username:
            info['hostname'] = platform.node()
            info['username'] = getpass.getuser()
            info['workdir'] = os.getcwd()

        return info

    def read_ini(self, option, option_type, default=configparser._UNSET):
        """
        Read option from the ini file for this sampler
        and also save to the output file if it exists
        """
        if default is None:
            default = configparser._UNSET
        if option_type is float:
            val = self.ini.getfloat(self.name, option, fallback=default)
        elif option_type is int:
            val = self.ini.getint(self.name, option, fallback=default)
        elif option_type is bool:
            val = self.ini.getboolean(self.name, option, fallback=default)
        elif option_type is str:
            val = self.ini.get(self.name, option, fallback=default)
        else:
            raise ValueError("Internal cosmosis sampler error: "
                "tried to read ini file option with unknown type {}".format(option_type))
        if self.output:
            self.output.metadata(option, str(val))
        return val

    def read_ini_choices(self, option, option_type, choices, default=None):
        value = self.read_ini(option, option_type, default=default)
        if value not in choices:
            name = self.__class__.__name__
            raise ValueError("Parameter {} for sampler {} must be one of: {}\n Parameter file said: {}".format(option, name, choices, value))
        return value


    def config(self):
        ''' Set up sampler (could instead use __init__) '''
        pass

    def execute(self):
        ''' Run one (self-determined) iteration of sampler.
            Should be enough to test convergence '''
        raise NotImplementedError

    def write_resume_info(self, info):
        try:
            filename = self.output.name_for_sampler_resume_info()
        except NotImplementedError:
            return

        # in some fast pipelines like demo 5 a keyboard interrupt
        # is likely to happen in the middle of this dump operation
        tmp_filename = filename + '.tmp'

        with open(tmp_filename, 'wb') as f:
            pickle.dump(info, f)

        shutil.move(tmp_filename, filename)

    def read_resume_info(self):
        filename = self.output.name_for_sampler_resume_info()
        if not os.path.exists(filename):
            return None
        with open(filename, 'rb') as f:
            return pickle.load(f)

    @classmethod
    def get_sampler(cls, name):
        try:
            return cls.registry[name.lower()]
        except KeyError:
            raise KeyError(f"Unknown sampler {name}")


    def resume(self):
        raise NotImplementedError("The sampler {} does not support resuming".format(self.name))

    def is_converged(self):
        return False
    


    def start_estimate(self, method=None, input_source=None, prefer_random=False):
        """
        Select a starting parameter set for the sampler.

        The method is chosen by looking at the start_method and start_input
        options in the sampler's ini file.

        This can be:
        - the peak from a previous sampler (default if available)
        - the defined starting position in the values file (if start_method not set)
        - a random point in the prior (if start_method is "prior")
        - a random point from a chain file (if start_method "chain-sample")
        - the last point from a chain file (if start_method is "chain-last")
        - a point from a covariance matrix (if start_method is "cov")

        Returns
        -------
        start : np.ndarray
        """
        if method is None:
            method = self.read_ini("start_method", str, "")
        if input_source is None:
            input_source = self.read_ini("start_input", str, "")

        if method == "chain":
            if prefer_random:
                method = "chain-sample"
            else:
                method = "chain-last"

        if self.distribution_hints.has_peak():
            start = self.distribution_hints.get_peak()
            if prefer_random:
                covmat = self.distribution_hints.get_cov()
                start = sample_ellipsoid(start, covmat)

        elif method == "prior":
            start = self.pipeline.randomized_start()

        elif method == "chain-last":
            if not input_source:
                raise ValueError("If you set the start_method to 'chain' you should not also set start_input to the name of a chain file")
            start = np.genfromtxt(input_source, invalid_raise=False)[-1, :self.pipeline.nvaried]

        elif method == "chain-sample":
            if not input_source:
                raise ValueError("If you set the start_method to 'chain' you should not also set start_input to the name of a chain file")

            # assume the chain file has a header. check for weight or log_weight columns
            # otherwise assume that the columns match the varied parameters
            data = np.genfromtxt(input_source, invalid_raise=False)
            with open(input_source) as f:
                maybe_colnames = f.readline().strip('#').split()
            if 'weight' in maybe_colnames:
                weight_index = maybe_colnames.index('weight')
                weight = data[:, weight_index]
            elif 'log_weight' in maybe_colnames:
                log_weight_index = maybe_colnames.index('log_weight')
                weight = np.exp(data[:, log_weight_index] - data[:, log_weight_index].max())
            else:
                weight = np.ones(len(data))
            index = np.random.choice(len(data), p=weight/weight.sum())
            start = data[index, :self.pipeline.nvaried]

        elif method == "cov":
            if not input_source:
                raise ValueError("If you set the start_method to 'cov' you should not also set start_input to the name of a covariance file")

            covmat = np.loadtxt(input_source)[:self.pipeline.nvaried, :self.pipeline.nvaried]
            start = sample_ellipsoid(self.pipeline.start_vector(), covmat)[0]

        else:
            # default method is just to use a single starting point
            start = self.pipeline.start_vector()
        return start

    def cov_estimate(self):
        covmat_file = self.read_ini("covmat", str, "")
        n = len(self.pipeline.varied_params)

        if self.distribution_hints.has_cov():
            # hints from a previous sampler take
            # precendence
            covmat = self.distribution_hints.get_cov()

        elif covmat_file:
            covmat = np.loadtxt(covmat_file)
            # Fix the size.
            # If there is only one sample parameter then 
            # we assume it is a 1x1 matrix
            # If it's a 1D vector then assume these are
            # standard deviations
            if covmat.ndim == 0:
                covmat = covmat.reshape((1, 1))
            elif covmat.ndim == 1:
                covmat = np.diag(covmat ** 2)

            # Error on incorrect shapes or sizes
            if covmat.shape[0] != covmat.shape[1]:
                raise ValueError("Covariance matrix from {}"
                                 "not square".format(covmat_file))
            if covmat.shape[0] != n:
                raise ValueError("Covariance matrix from {} "
                                 "is the wrong shape ({}x{}) "
                                 "for the {} varied params".format(
                                    covmat_file, covmat.shape[0], n))
        else:
            # Just try a minimal estimate - 5% of prior width as standard deviation
            covmat_scale = self.read_ini("covmat_scale", float, 0.05)
            covmat = np.diag([covmat_scale*p.width() for p in self.pipeline.varied_params])**2

        return covmat



class ParallelSampler(Sampler):
    parallel_output = True
    is_parallel_sampler = True
    supports_smp = True
    def __init__(self, ini, pipeline, output=None, pool=None):
        Sampler.__init__(self, ini, pipeline, output)
        self.pool = pool

    def worker(self):
        ''' Default to a map-style worker '''
        if self.pool:
            self.pool.wait()
        else:
            raise RuntimeError("Worker function called when no parallel pool exists!")

    def is_master(self):
        return self.pool is None or self.pool.is_master()


# These are marked as deprecated in emcee, so I moved them here.
# I think I wrote the first one.  And I've rewritten the second
# to use the first
def sample_ellipsoid(p0, covmat, size=1):
    """
    Produce an ellipsoid of walkers around an initial parameter value,
    according to a covariance matrix.
    :param p0: The initial parameter value.
    :param covmat:
        The covariance matrix.  Must be symmetric-positive definite or
        it will raise the exception numpy.linalg.LinAlgError
    :param size: The number of samples to produce.
    """
    return np.random.multivariate_normal(
        np.atleast_1d(p0), np.atleast_2d(covmat), size=size
    )

def sample_ball(p0, std, size=1):
    """
    Produce a ball of walkers around an initial parameter value.
    :param p0: The initial parameter value.
    :param std: The axis-aligned standard deviation.
    :param size: The number of samples to produce.
    """
    covmat = np.diag(std**2)
    return sample_ellipsoid(p0, covmat, size)



G           =  6.673840e-11  # m3 / (kg s2)  # Gravitational constant
L_sun       =  3.846000e+26  # W             # Solar luminosity
M_earth     =  5.974200e+24  # kg            # Earth mass
M_jup       =  1.898700e+27  # kg            # Jupiter mass
M_sun       =  1.989100e+30  # kg            # Solar mass
N_A         =  6.022141e+23  # 1 / mol       # Avogadro's number
R           =  8.314462e+00  # J / (K mol)   # Gas constant
R_earth     =  6.378136e+06  # m             # Earth equatorial radius
R_jup       =  7.149200e+07  # m             # Jupiter equatorial radius
R_sun       =  6.955080e+08  # m             # Solar radius
Ryd         =  1.097373e+07  # 1 / m         # Rydberg constant
a0          =  5.291772e-11  # m             # Bohr radius
alpha       =  7.297353e-03  #               # Fine-structure constant
atmosphere  =  1.013250e+05  # Pa            # Atmosphere
au          =  1.495979e+11  # m             # Astronomical Unit
b_wien      =  2.897772e-03  # K m           # Wien wavelength displacement law constant
c           =  2.997925e+08  # m / s         # Speed of light in vacuum
e           =  1.602177e-19  # C             # Electron charge
eps0        =  8.854188e-12  # F / m         # Electric constant
g0          =  9.806650e+00  # m / s2        # Standard acceleration of gravity
h           =  6.626070e-34  # J s           # Planck constant
hbar        =  1.054572e-34  # J s           # Reduced Planck constant
k_B         =  1.380649e-23  # J / K         # Boltzmann constant
kpc         =  3.085678e+19  # m             # Kiloparsec
m_e         =  9.109383e-31  # kg            # Electron mass
m_n         =  1.674927e-27  # kg            # Neutron mass
m_p         =  1.672622e-27  # kg            # Proton mass
mu0         =  1.256637e-06  # N / A2        # Magnetic constant
muB         =  9.274010e-24  # J / T         # Bohr magneton
pc          =  3.085678e+16  # m             # Parsec
sigma_sb    =  5.670373e-08  # W / (K4 m2)   # Stefan-Boltzmann constant
u           =  1.660539e-27  # kg            # Atomic mass




import logging
import sys

NOISY = 15


# Change the logger so it doesn't print out all the boilerplate, unless requested
logger = logging.getLogger("cosmosis")
handler = logging.StreamHandler(sys.stdout)
logger.addHandler(handler)
formatter = logging.Formatter('%(message)s')
handler.setFormatter(formatter)
logger.propagate = False


verbosity_levels = {
    "debug": 40,    #40 log level debug
    "noisy": 35,               #35 log level noisy
    "standard": 30,  #30 log level info
    "normal": 30,  #30 log level info
    "quiet": 20, #20 log level warning / overview
    "muted": 10,    #10 log level error / important
    "silent": -1,   #-1
}


def set_verbosity(verb):
    try:
        verb = int(verb)
    except ValueError:
        pass
    if not isinstance(verb, int):
        try:
            verb = verbosity_levels[verb]
        except KeyError:
            valid_levels = ', '.join(list(verbosity_levels.keys()))
            message = """Error specifiying verbosity.
                You put: '{0}'.
                We want either an integer 0 (silent) - 50 (everything) 
                or one of: {1}""".format(verb, valid_levels)
            raise ValueError(message)
    level = 50 - verb
    debug(f"CosmoSIS verbosity set to {verb}")
    set_level(level)

def set_level(level):
    logger.setLevel(level)
    handler.setLevel(level)

def get_level():
    return logger.getEffectiveLevel()

def enable_log_tracing():
    formatter = logging.Formatter('[%(pathname)s Line %(lineno)d]: %(message)s')
    handler.setFormatter(formatter)

# level 10
def debug(message):
    logger.log(logging.DEBUG, message)
# extreme level debugging information, printing basically everything

# level 15
def noisy(message):
    logger.log(NOISY, message)

# level 20
def info(message):
    logger.log(logging.INFO, message)

# level 30
def warning(message):
    logger.log(logging.WARNING, message)

def overview(message):
    logger.log(logging.WARNING, message)

# Level 40
def error(message):
    logger.log(logging.ERROR, message)

def important(message):
    logger.log(logging.ERROR, message)

# Level 50
def critical(message):
    logger.log(logging.CRITICAL, message)

def is_enabled_for(level):
    return logger.isEnabledFor(level)



import setuptools
import setuptools.command.install
import setuptools.command.build_py
import setuptools.command.develop
import subprocess
import os

version = open('cosmosis/version.py').read().split('=')[1].strip().strip("'")


f90_mods = [
    "datablock/cosmosis_section_names.mod",
    "datablock/cosmosis_types.mod",
    "datablock/cosmosis_wrappers.mod",
    "datablock/cosmosis_modules.mod",
]

scripts = [
    'bin/cosmosis',
    'bin/cosmosis-campaign',
    'bin/cosmosis-configure',
    'bin/cosmosis-extract',
    'bin/cosmosis-sample-fisher',
    'bin/cosmosis-postprocess',
]

c_headers = [
    "datablock/c_datablock.h",
    "datablock/datablock_logging.h",
    "datablock/datablock_types.h",
    "datablock/cosmosis_constants.h",
    "datablock/datablock_status.h",
    "datablock/section_names.h",
]

cc_headers = [
    "datablock/clamp.hh",
    "datablock/entry.hh",
    "datablock/fakearray.hh",
    "datablock/ndarray.hh",
    "datablock/datablock.hh",
    "datablock/exceptions.hh",
    "datablock/mdarraygen.hh",
    "datablock/section.hh"
]

datablock_libs = ["datablock/libcosmosis.so"]

sampler_libs = ["samplers/multinest/multinest_src/libnest3.so",
                "samplers/multinest/multinest_src/libnest3_mpi.so",
                "samplers/polychord/polychord_src/libchord.so",
                "samplers/polychord/polychord_src/libchord_mpi.so",
                "samplers/minuit/minuit_wrapper.so"]


testing_files = [
    "test/libtest/c_datablock_complex_array_test.c",
    "test/libtest/c_datablock_double_array_test.c",
    "test/libtest/c_datablock_int_array_test.c",
    "test/libtest/c_datablock_multidim_complex_array_test.c",
    "test/libtest/c_datablock_multidim_double_array_test.c",
    "test/libtest/c_datablock_multidim_int_array_test.c",
    "test/libtest/c_datablock_test.c",
    "test/libtest/cosmosis_test.F90",
    "test/libtest/cosmosis_tests.supp",
    "test/libtest/datablock_test.cc",
    "test/libtest/entry_test.cc",
    "test/libtest/ndarray_test.cc",
    "test/libtest/section_test.cc",
    "test/libtest/test_c_datablock_scalars.h",
    "test/libtest/test_c_datablock_scalars.template",
    "test/libtest/Makefile",
    "test/campaign.yml",
    "test/included.yml",
    "test/bad-campaign.yml",
    "test/example-priors.ini",
    "test/example-values.ini",
    "test/example.ini",

]

other_files = ["postprocessing/latex.ini"]

compilers_config = ["config/compilers.mk", "config/subdirs.mk"]


def get_COSMOSIS_SRC_DIR():
    cosmosis_src_dir = os.path.join(os.getcwd(), "cosmosis")
    return cosmosis_src_dir

def make_cosmosis():
    print("Running CosmoSIS main library compile command")
    cosmosis_src_dir = get_COSMOSIS_SRC_DIR()
    env = os.environ.copy()
    env["COSMOSIS_SRC_DIR"] = cosmosis_src_dir

    # If we are in a conda build env then the appropriate
    # variable is called PREFIX. Otherwise if we are on
    # a user's build env then it is called CONDA_PREFIX
    if os.environ.get("CONDA_BUILD", "0") == "1":
        prefix_name = "PREFIX"
        conda_build = True
    else:
        prefix_name = "CONDA_PREFIX"
        conda_build = False

    conda = env.get(prefix_name)
    if conda:
        # User can switch on COSMOSIS_OMP manually, but it should
        # always be on for conda.
        env["COSMOSIS_OMP"] = "1"

        # We also need Lapack to build some of the samplers
        env["LAPACK_LINK"] = f"-L{conda}/lib -llapack"
        # and minuit for that sampler
        env['MINUIT2_LIB'] = f"{conda}/lib"
        env['MINUIT2_INC'] = f"{conda}/include/Minuit2"

        if conda_build:
            env["COSMOSIS_ALT_COMPILERS"] = "1"
            env['USER_FFLAGS'] = env['FFLAGS']
            env['USER_CFLAGS'] = env['CFLAGS']
            env['USER_CXXFLAGS'] = env['CXXFLAGS']

        # and the MPI compiler
        env["MPIFC"] = "mpif90"

    env['FC'] = env.get('FC', 'gfortran')

    subprocess.check_call(["make"], env=env, cwd="cosmosis")

class build_cosmosis(setuptools.Command):
    description = "Build CosmoSIS and do nothing else"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        make_cosmosis()



class build_py_cosmosis(setuptools.command.build_py.build_py):
    def run(self):
        make_cosmosis()
        super().run()

class clean_cosmosis(setuptools.Command):
    description = "Run the CosmoSIS clean process"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Running CosmoSIS main library clean command")
        cosmosis_src_dir = get_COSMOSIS_SRC_DIR()
        env = {"COSMOSIS_SRC_DIR": cosmosis_src_dir,}
        subprocess.check_call(["make", "clean"], env=env, cwd="cosmosis")

class install_cosmosis(setuptools.command.install.install):
    description = "Run the CosmoSIS install process"

    def run(self):
        make_cosmosis()
        super().run()

class develop_cosmosis(setuptools.command.develop.develop):
    description = "Install CosmoSIS in editable mode"
    def run(self):
        make_cosmosis()
        super().run()


requirements = [
    "pyyaml",
    "emcee",
    "numpy",
    "scipy",
    "matplotlib",
    "pybind11",
    "pyyaml",
    "scipy",
    "threadpoolctl",
    "emcee",
    "dynesty",
    "zeus-mcmc",
    "nautilus-sampler>=1.0.1",
    "dulwich",
    "scikit-learn",
    "future",
    "Py-BOBYQA",

]

all_package_files = (datablock_libs + sampler_libs
                            + c_headers + cc_headers + f90_mods 
                            + compilers_config + testing_files + other_files)

setuptools.setup(name = 'cosmosis',
    description       = "The CosmoSIS parameter estimation library.",
    author            = "Joe Zuntz",
    author_email      = "joezuntz@googlemail.com",
    url               = "https://github.com/joezuntz/cosmosis",
    packages = setuptools.find_packages(),
    package_data = {"cosmosis" : all_package_files},
    scripts = scripts,
    install_requires = requirements,
    cmdclass={
        "build_cosmosis": build_cosmosis,
        "build_py": build_py_cosmosis,
        "install": install_cosmosis,
        "develop": develop_cosmosis,
        "clean": clean_cosmosis,
    },
    version=version,
)




from .main import run_cosmosis
from .runtime import MPIPool, LikelihoodPipeline, FunctionModule, \
                     stdout_redirected, Inifile, CosmosisConfigurationError, \
                     MPIPool, Module, FunctionModule, ClassModule, logs
from .samplers import Sampler
from . import samplers
from .version import __version__
from .datablock import option_section, DataBlock
from .gaussian_likelihood import GaussianLikelihood, \
                                 SingleValueGaussianLikelihood, \
                                 WindowedGaussianLikelihood
from .output import TextColumnOutput, InMemoryOutput, NullOutput, AstropyOutput, CosmoMCOutput, FitsOutput



from .. import Sampler
import numpy as np
from ...runtime import pipeline
import sys


class TestSampler(Sampler):
    # The name of this class makes pytest think that it is a
    # test to be run, instead of a tool for testing likelhioods.
    # This should stop that from happening.
    __test__ = False

    needs_output = False

    def config(self):
        self.converged = False
        self.fatal_errors = self.read_ini("fatal_errors", bool, False)
        self.save_dir = self.read_ini("save_dir", str, "")
        self.graph = self.read_ini("graph", str, "")
        self.analyze_fast_slow = self.read_ini("analyze_fast_slow", bool, False)
        self.print_log = self.read_ini("print_log", bool, False)

    def execute(self):
        # load initial parameter values
        p = np.array([param.start for param in self.pipeline.varied_params])
    
        # try to print likelihood if it exists
        data=None
        try:
            prior = self.pipeline.prior(p)
            like, extra, data = self.pipeline.likelihood(p, return_data=True)
            if self.pipeline.likelihood_names:
                print("Prior      = ", prior)
                print("Likelihood = ", like)
                print("Posterior  = ", like+prior)
        except pipeline.MissingLikelihoodError as error:
            found_likelihoods = [k[1][:-5] for k in list(error.pipeline_data.keys()) if k[0]=="likelihoods"]            
            sys.stderr.write("\n")
            sys.stderr.write("One of the likelihoods you asked for was not found.\n")
            sys.stderr.write("You asked for: %s\n"%str(error))
            sys.stderr.write("But the only ones calculated in the pipeline were:\n")
            sys.stderr.write(", ".join(found_likelihoods)+"\n")
            sys.stderr.write("\n")
            if self.fatal_errors:
                raise
        except Exception as e:
            if self.fatal_errors:
                raise
            print("(Could not get a likelihood) Error:"+str(e))
        if not self.pipeline.likelihood_names:
            print("(No likelihoods required in ini file)")
            print()

        if self.analyze_fast_slow:
            print("\n")
            self.pipeline.do_fast_slow = True
            self.pipeline.setup_fast_subspaces(all_params=True)
            print("\n")


        if self.graph:
            self.pipeline.make_graph(data, self.graph)

        try:
            if self.save_dir:
                if data is not None:
                    if self.save_dir.endswith('.tgz'):
                        data.save_to_file(self.save_dir[:-4], clobber=True)
                    else:
                        data.save_to_directory(self.save_dir, clobber=True)
                else:
                    print("(There was an error so no output to save)")
        except Exception as e:
            if self.fatal_errors:
                raise
            print("Could not save output.")

        if data is None and self.fatal_errors:
            raise RuntimeError("Pipeline failed at some stage")

        if self.print_log and data is not None:
            data.print_log()


        self.converged = True

    def is_converged(self):
        return self.converged
